from ..doctools import document
from .geom_area import geom_area


@document
class geom_density(geom_area):
    """
    Smooth density estimate

    {usage}

    Parameters
    ----------
    {common_parameters}

    See Also
    --------
    plotnine.geom_ribbon
    """

    DEFAULT_AES = {
        **geom_area.DEFAULT_AES,
        "color": "black",
        "fill": None,
        "weight": 1,
    }

    DEFAULT_PARAMS = {
        **geom_area.DEFAULT_PARAMS,
        "stat": "density",
        "position": "identity",
    }



from __future__ import annotations

from typing import TYPE_CHECKING

from matplotlib import artist
from matplotlib.patches import FancyBboxPatch, Rectangle
from matplotlib.text import _get_textbox  # type: ignore
from matplotlib.transforms import Affine2D

if TYPE_CHECKING:
    from matplotlib.backend_bases import RendererBase

    from plotnine.typing import StripPosition

    from .text import StripText


# We subclass because we want to learn the size and location
# of the box when the layout manager is running.
# With MPLs default text & boxpatch, the patch gets its
# dimension information at draw time.


class StripTextPatch(FancyBboxPatch):
    """
    Strip text background box
    """

    # The text artists that is wrapped by this box
    text: StripText
    position: StripPosition
    _update = False

    def __init__(self, text: StripText):
        boxstyle = f"square, pad={text.draw_info.strip_text_margin}"

        super().__init__(
            (0, 0), 1, 1, boxstyle=boxstyle, clip_on=False, zorder=2.2
        )

        self.text = text
        self.position = text.draw_info.position

    def update_position_size(self, renderer: RendererBase):
        """
        Update the location and the size of the bbox.
        """
        if self._update:
            return

        text = self.text
        posx, posy = text.get_transform().transform((text._x, text._y))
        x, y, w, h = _get_textbox(text, renderer)

        self.set_bounds(0.0, 0.0, w, h)
        self.set_transform(
            Affine2D()
            .rotate_deg(text.get_rotation())
            .translate(posx + x, posy + y)
        )
        fontsize_in_pixel = renderer.points_to_pixels(
            text.get_size()  # type: ignore
        )
        self.set_mutation_scale(fontsize_in_pixel)  # type: ignore
        self._update = True

    def get_window_extent(self, renderer=None):
        """
        Location & dimensions of the box
        """
        if renderer:
            self.update_position_size(renderer)
        return super().get_window_extent(renderer)


class InsideStrokedRectangle(Rectangle):
    """
    A rectangle whose stroked is fully contained within it
    """

    @artist.allow_rasterization
    def draw(self, renderer):
        """
        Draw with the bounds of the rectangle adjusted to accomodate the stroke
        """
        x, y = self.xy
        w, h = self.get_width(), self.get_height()
        lw = self.get_linewidth()
        self.set_bounds((x + lw / 2), (y + lw / 2), (w - lw), (h - lw))
        super().draw(renderer)



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from typing import Any, Literal, Optional, Type

    from plotnine import theme
    from plotnine.typing import FigureFormat

close_all_figures = False
"""
Development flag, e.g. set to `True` to prevent
the queuing up of figures when errors happen.
"""

current_theme: Optional[theme | Type[theme]] = None
"""Theme used when none is added to the ggplot object"""

base_family: str = "sans-serif"
"""
The base font family for all text that is part of the theme.
Default is sans-serif and one is choosen automatically from
rcParams["font.san-serif"]
"""

aspect_ratio: Literal["auto"] | float = "auto"
"""
Default aspect ratio used by the themes
"""

dpi: int = 100
"""
Default DPI used by the themes
"""

figure_size: tuple[float, float] = (640 / dpi, 480 / dpi)
"""
Default figure size inches
"""

figure_format: Optional[FigureFormat] = None
"""
The format for the inline figures outputed by the jupyter kernel.

If `None`, it is the value of

    %config InlineBackend.figure_format

If that has not been set, the default is "retina".
You can set it explicitly with:

    %config InlineBackend.figure_format = "retina"
"""

base_margin: float = 0.01
"""
A size that is proportional of the figure width and
is used by some themes to determine other margins
"""


def get_option(name: str) -> Any:
    """
    Get package option

    Parameters
    ----------
    name :
        Name of the option
    """
    d = globals()

    if name in {"get_option", "set_option"} or name not in d:
        from .exceptions import PlotnineError

        raise PlotnineError(f"Unknown option {name}")

    return d[name]


def set_option(name: str, value: Any) -> Any:
    """
    Set package option

    Parameters
    ----------
    name :
        Name of the option
    value :
        New value of the option

    Returns
    -------
    :
        Old value of the option
    """
    d = globals()

    if name in {"get_option", "set_option"} or name not in d:
        from .exceptions import PlotnineError

        raise PlotnineError(f"Unknown option {name}")

    old = d[name]
    d[name] = value
    return old



"""
Plotnine Data
"""

from pathlib import Path

import pandas as pd
from pandas.api.types import CategoricalDtype

__all__ = (
    "diamonds",
    "economics",
    "economics_long",
    "faithful",
    "faithfuld",
    "huron",
    "luv_colours",
    "meat",
    "midwest",
    "mpg",
    "msleep",
    "mtcars",
    "pageviews",
    "penguins",
    "presidential",
    "seals",
    "txhousing",
)

DATA_DIR = Path(__file__).parent

mtcars = pd.read_csv(DATA_DIR / "mtcars.csv")
meat = pd.read_csv(DATA_DIR / "meat.csv", parse_dates=[0])
pageviews = pd.read_csv(DATA_DIR / "pageviews.csv", parse_dates=[0])
huron = pd.read_csv(DATA_DIR / "huron.csv")
seals = pd.read_csv(DATA_DIR / "seals.csv")
economics = pd.read_csv(DATA_DIR / "economics.csv", parse_dates=[0])
economics_long = pd.read_csv(DATA_DIR / "economics_long.csv", parse_dates=[0])
presidential = pd.read_csv(DATA_DIR / "presidential.csv", parse_dates=[1, 2])
txhousing = pd.read_csv(DATA_DIR / "txhousing.csv")
penguins = pd.read_csv(DATA_DIR / "penguins.csv")
luv_colours = pd.read_csv(DATA_DIR / "luv_colours.csv")
faithfuld = pd.read_csv(DATA_DIR / "faithfuld.csv")
faithful = pd.read_csv(DATA_DIR / "faithful.csv")

# For conveniece to the user, we set some columns in these
# dataframes to categoricals.
diamonds = pd.read_csv(DATA_DIR / "diamonds.csv")
midwest = pd.read_csv(DATA_DIR / "midwest.csv")
mpg = pd.read_csv(DATA_DIR / "mpg.csv")
msleep = pd.read_csv(DATA_DIR / "msleep.csv")


# add factors
def _ordered_categories(df, categories):
    """
    Make the columns in df categorical

    Parameters
    ----------
    df : pd.Dataframe
        Dataframe
    categories: dict
        Of the form {str: list},
        where the key the column name and the value is
        the ordered category list
    """
    for col, cats in categories.items():
        df[col] = df[col].astype(CategoricalDtype(cats, ordered=True))
    return df


def _unordered_categories(df, columns):
    """Make the columns in df categorical"""
    for col in columns:
        df[col] = df[col].astype(CategoricalDtype(ordered=False))
    return df


def _process_categories():
    """
    Set columns in some of the dataframes to categoricals
    """
    global diamonds, midwest, mpg, msleep, penguins
    diamonds = _ordered_categories(
        diamonds,
        {
            "cut": "Fair, Good, Very Good, Premium, Ideal".split(", "),
            "clarity": "I1 SI2 SI1 VS2 VS1 VVS2 VVS1 IF".split(),
            "color": list("DEFGHIJ"),
        },
    )
    mpg = _unordered_categories(
        mpg, "manufacturer model trans fl drv class".split()
    )
    midwest = _unordered_categories(midwest, ["category"])
    msleep = _unordered_categories(msleep, ["vore", "conservation"])
    penguins = _unordered_categories(penguins, ["species", "island", "sex"])


_process_categories()


# Documentation

mtcars.__doc__ = """
Motor Trend Car Road Tests

## Description

The data was extracted from the 1974 *Motor Trend* US magazine,
and comprises fuel consumption and 10 aspects of automobile
design and performance for 32 automobiles (1973–74 models).

## Format

A data frame with 32 observations on 11 variables.

| Column | Description                              |
|--------|------------------------------------------|
| mpg    | Miles/(US) gallon                        |
| cyl    | Number of cylinders                      |
| disp   | Displacement (cu.in.)                    |
| hp     | Gross horsepower                         |
| drat   | Rear axle ratio                          |
| wt     | Weight (1000 lbs)                        |
| qsec   | 1/4 mile time                            |
| vs     | V/S                                      |
| am     | Transmission (0 = automatic, 1 = manual) |
| gear   | Number of forward gears                  |
| carb   | Number of carburetors                    |

## Source

Henderson and Velleman (1981), Building multiple regression
models interactively. *Biometrics*, **37**, 391–411.
"""

meat.__doc__ = """
US Meat Production

Description**

Monthly production weight of red meat and poultry. The values are in
units of 1 million lbs.

| Column            | Description           |
|-------------------|-----------------------|
| date              | Month of the year     |
| beef              | Beef weight           |
| veal              | Veal weight           |
| pork              | Pork weight           |
| lamb_and_mutton   | Lamb & Mutton weight  |
| broilers          | Broiler weight        |
| other_chicken     | Other Chicken weight  |
| turkey            | Turkey weight         |

## Source

[Livestock and Meat Dosmestic Data]\
(https://www.ers.usda.gov/data-products/livestock-and-meat-domestic-data/)
from the Economic Research Service of the U.S. DEPARTMENT OF AGRICULTURE.

"""

pageviews.__doc__ = """
"""

huron.__doc__ = """
Level of Lake Huron 1875–1972

Description**

Annual measurements of the level, in feet, of Lake Huron 1875–1972.

## Format

| Column | Description |
|--------|-------------|
| year   | Year        |
| level  | Water level |
| decade | Decade      |

Source**

Brockwell, P. J. and Davis, R. A. (1991). Time Series and Forecasting Methods.
Second edition. Springer, New York. Series A, page 555.

Brockwell, P. J. and Davis, R. A. (1996). Introduction to Time Series and
Forecasting. Springer, New York. Sections 5.1 and 7.6.
"""

seals.__doc__ = """
Vector field of seal movements.

## Description

This vector field was produced from the data described in Brillinger,
D.R., Preisler, H.K., Ager, A.A. and Kie, J.G.
"An exploratory data analysis (EDA) of the paths of moving animals". J.
Statistical Planning and Inference 122 (2004), 43-63, using the methods
of Brillinger, D.R., "Learning a potential function from a trajectory",
Signal Processing Letters. December (2007).

## Format

A data frame with 1155 rows and 4 variables

| Column     | Description         |
|------------|---------------------|
| lat        | Latitude            |
| long       | Longitude           |
| delta_long | Change in Longitude |
| delta_lat  | Change in Latitude  |

## References

<http://www.stat.berkeley.edu/~brill/Papers/jspifinal.pdf>
"""

economics.__doc__ = """
US economic time series.

## Description

This dataset was produced from US economic time series data available
from http://research.stlouisfed.org/fred2.
`economics` is in "wide" format, `economics_long` is in "long" format.

## Format

A data frame with 478 rows and 6 variables

| Column   | Description                                                    |
|----------|----------------------------------------------------------------|
| date     | Month of data collection                                       |
| psavert  | personal savings rate [^1]                                     |
| pce      | personal consumption expenditures, in billions of dollars [^2] |
| unemploy | number of unemployed in thousands [^3]                         |
| uempmed  | median duration of unemployment, in week [^4]                  |
| pop      | total population, in thousands [^5]                            |

[^1]: <http://research.stlouisfed.org/fred2/series/PSAVERT>
[^2]: <http://research.stlouisfed.org/fred2/series/PCE>
[^3]: <http://research.stlouisfed.org/fred2/series/UNEMPLOY>
[^4]: <http://research.stlouisfed.org/fred2/series/UEMPMED>
[^5]: <http://research.stlouisfed.org/fred2/series/POP>
"""

economics_long.__doc__ = economics.__doc__

penguins.__doc__ = """
Palmer Penguins

## Description

Data about 3 different species of penguins collected from 3 islands
in the Palmer Archipelago, Antarctica.

## Format

A dataframe with 344 rows and 8 variables

+--------------------+------------------------------------------------+
| Column             | Description                                    |
+====================+================================================+
| species            | Penguin species (Adélie, Chinstrap and Gentoo) |
+--------------------+------------------------------------------------+
| island             | Island in Palmer Archipelago, Antarctica       |
|                    | (Biscoe, Dream or Torgersen)                   |
+--------------------+------------------------------------------------+
| bill_length_mm     | Bill length (millimeters)                      |
+--------------------+------------------------------------------------+
| bill_depth_mm      | Bill depth (millimeters)                       |
+--------------------+------------------------------------------------+
| flipper_length_mm  | Flipper length (millimeters)                   |
+--------------------+------------------------------------------------+
| body_mass_g        | Body mass (grams)                              |
+--------------------+------------------------------------------------+
| sex                | Penguin sex (female, male)                     |
+--------------------+------------------------------------------------+
| year               | The study year (2007, 2008, or 2009)           |
+--------------------+------------------------------------------------+

## Source

Collected by [Dr. Kristen Gorman]\
(https://www.uaf.edu/cfos/people/faculty/detail/kristen-gorman.php)
and the [Palmer Station, Antarctica LTER](https://pallter.marine.rutgers.edu/).

Made conveniently availabe by
[Alison Horst](https://github.com/allisonhorst/palmerpenguins/) to serve as
a dataset exploration and visualisation.
"""

presidential.__doc__ = """
Terms of 11 presidents from Eisenhower to Obama.

## Description

The names of each president, the start and end date
of their term, and their party of 11 US presidents
from Eisenhower to Obama.

## Format

A data frame with 11 rows and 4 variables

| Column | Description                |
|--------|----------------------------|
| name   | Name of president          |
| start  | Start of presidential term |
| end    | End of presidential term   |
| party  | Political Party            |

"""

txhousing.__doc__ = """
Housing sales in TX.

## Description

Information about the housing market in Texas provided
by the TAMU real estate center, http://recenter.tamu.edu/.

## Format

A data frame with 8602 observations and 9 variables:

+-----------+-----------------------------------------------+
| Column    | Description                                   |
+===========+===============================================+
| city      | Name of MLS area                              |
+-----------+-----------------------------------------------+
| year      | Year                                          |
+-----------+-----------------------------------------------+
| month     | Month                                         |
+-----------+-----------------------------------------------+
| sales     | Number of sales                               |
+-----------+-----------------------------------------------+
| volume    | Total value of sales                          |
+-----------+-----------------------------------------------+
| median    | Median sale price                             |
+-----------+-----------------------------------------------+
| listings  | Total active listings                         |
+-----------+-----------------------------------------------+
| inventory | "Months inventory": amount of time it would   |
|           | take to sell all current listings at current  |
|           | pace of sales.                                |
+-----------+-----------------------------------------------+
| date      | Date                                          |
+-----------+-----------------------------------------------+

"""

luv_colours.__doc__ = """
colors in Luv space.

## Description

Named colors translated into Luv colour space.

## Format

A data frame with 657 observations and 4 variables:

| Column | Description                    |
|--------|--------------------------------|
| L      | L position in Luv colour space |
| u      | u position in Luv colour space |
| v      | v position in Luv colour space |
| col    | Colour name                    |
"""

faithful.__doc__ = """
Old Faithful Geyser Data

## Description

Waiting time between eruptions and the duration of the
eruption for the Old Faithful geyser in Yellowstone
National Park, Wyoming, USA.

## Format

A data frame with 272 observations on 2 variables.

| Column    | Description                            |
|-----------|----------------------------------------|
| eruptions | Eruption time in mins                  |
| waiting W | aiting time to next eruption (in mins) |

## Details

A closer look at `faithful.eruptions` reveals that these are
heavily rounded times originally in seconds, where multiples
of 5 are more frequent than expected under non-human measurement.
For a better version of the eruption times, see the example below.

There are many versions of this dataset around:
Azzalini and Bowman (1990) use a more complete version.

## Source

W. Härdle.

## References

Härdle, W. (1991) *Smoothing Techniques with Implementation in S*.
New York: Springer.

Azzalini, A. and Bowman, A. W. (1990). A look at some data
on the Old Faithful geyser. **Applied Statistics** *39*, 357–365.
"""

faithfuld.__doc__ = """
Old Faithful Geyser Data

## Description

Waiting time between eruptions and the duration of the
eruption for the Old Faithful geyser in Yellowstone
National Park, Wyoming, USA.

## Format

A data frame with *grid data* for 272 observations on 2
variables and the density at those locations.

| Column    | Description                            |
|-----------|----------------------------------------|
| eruptions | Eruption time in mins                  |
| waiting W | aiting time to next eruption (in mins) |
| density D | ensity Estimate                        |

## Details

A closer look at `faithful.eruptions` reveals that these are
heavily rounded times originally in seconds, where multiples
of 5 are more frequent than expected under non-human measurement.
For a better version of the eruption times, see the example below.

There are many versions of this dataset around:
Azzalini and Bowman (1990) use a more complete version.

## Source

W. Härdle.

## References

Härdle, W. (1991) *Smoothing Techniques with Implementation in S*.
New York: Springer.

Azzalini, A. and Bowman, A. W. (1990). A look at some data
on the Old Faithful geyser. **Applied Statistics** *39*, 357–365.
"""

diamonds.__doc__ = """
Prices of 50,000 round cut diamonds

## Description

A dataset containing the prices and other attributes
of almost 54,000 diamonds. The variables are as follows:

## Format

A data frame with 53940 rows and 10 variables:

+----------+------------------------------------------------------------+
| Column   | Description                                                |
+==========+============================================================+
| price    | price in US dollars ($326–$18,823)                         |
+----------+------------------------------------------------------------+
| carat    | weight of the diamond (0.2–5.01)                           |
+----------+------------------------------------------------------------+
| cut      | quality of the cut (Fair, Good, Very Good, Premium, Ideal) |
+----------+------------------------------------------------------------+
| color    | diamond colour, from J (worst) to D (best)                 |
+----------+------------------------------------------------------------+
| clarity  | a measurement of how clear the diamond is                  |
|          | (I1 (worst), SI1, SI2, VS1, VS2, VVS1, VVS2, IF (best))    |
+----------+------------------------------------------------------------+
| x        | length in mm (0–10.74)                                     |
+----------+------------------------------------------------------------+
| y        | width in mm (0–58.9)                                       |
+----------+------------------------------------------------------------+
| z        | depth in mm (0–31.8)                                       |
+----------+------------------------------------------------------------+
| depth    | total depth percentage                                     |
|          | = z / mean(x, y) = 2 * z / (x + y) (43–79)                 |
+----------+------------------------------------------------------------+
| table    | width of top of diamond relative to widest point (43–95)   |
+----------+------------------------------------------------------------+
"""

midwest.__doc__ = """
Midwest demographics.

## Description

Demographic information of midwest counties

## Format

A data frame with 437 rows and 28 variables

|Column                | Description                  |
|----------------------|------------------------------|
|PID                   |                              |
|county                |                              |
|state                 |                              |
|area                  |                              |
|poptotal              | Total population             |
|popdensity            | Population density           |
|popwhite              | Number of whites             |
|popblack              | Number of blacks             |
|popamerindian         | Number of American Indians   |
|popasian              | Number of Asians             |
|popother              | Number of other races        |
|percwhite             | Percent white                |
|percblack             | Percent black                |
|percamerindan         | Percent American Indian      |
|percasian             | Percent Asian                |
|percother             | Percent other races          |
|popadults             | Number of adults             |
|perchsd               |                              |
|percollege            | Percent college educated     |
|percprof              | Percent profession           |
|poppovertyknown       |                              |
|percpovertyknown      |                              |
|percbelowpoverty      |                              |
|percchildbelowpovert  |                              |
|percadultpoverty      |                              |
|percelderlypoverty    |                              |
|inmetro               | In a metro area              |
|category              |                              |
"""

mpg.__doc__ = """
Fuel economy data from 1999 and 2008 for 38 popular models of car

## Description

This dataset contains a subset of the fuel economy data that
the EPA makes available on http://fueleconomy.gov.
It contains only models which had a new release every year
between 1999 and 2008 - this was used as a proxy for
the popularity of the car.

## Format

A data frame with 234 rows and 11 variables

+--------------+---------------------------------+
| Column       | Description                     |
+==============+=================================+
| manufacturer |                                 |
+--------------+---------------------------------+
| model        |                                 |
+--------------+---------------------------------+
| displ        | engine displacement, in litres  |
+--------------+---------------------------------+
| year         |                                 |
+--------------+---------------------------------+
| cyl          | number of cylinders             |
+--------------+---------------------------------+
| trans        | type of transmission            |
+--------------+---------------------------------+
| drv          | f = front-wheel drive           |
|              | r = rear wheel drive            |
|              | 4 = 4wd                         |
+--------------+---------------------------------+
| cty          | city miles per gallon           |
+--------------+---------------------------------+
| hwy          | highway miles per gallon        |
+--------------+---------------------------------+
| fl           |                                 |
+--------------+---------------------------------+
| class        |                                 |
+--------------+---------------------------------+
"""

msleep.__doc__ = """
An updated and expanded version of the mammals sleep dataset.

## Description

This is an updated and expanded version of the mammals
sleep dataset. Updated sleep times and weights were taken
from V. M. Savage and G. B. West. A quantitative, theoretical
framework for understanding mammalian sleep. Proceedings of
the National Academy of Sciences, 104 (3):1051-1056, 2007.

## Format

A data frame with 83 rows and 11 variables

| Column       | Description                           |
|--------------|---------------------------------------|
| name genus   | common name                           |
| vore order   | carnivore, omnivore or herbivore?     |
| conservation | the conservation status of the animal |
| sleep_total  | total amount of sleep, in hours       |
| sleep_rem    | rem sleep, in hours                   |
| sleep_cycle  | length of sleep cycle, in hours       |
| awake        | amount of time spent awake, in hours  |
| brainwt      | brain weight in kilograms             |
| bodywt       | body weight in kilograms              |

## Details

Additional variables order, conservation status and
vore were added from wikipedia.
"""



from __future__ import annotations

import typing
from typing import Sized
from warnings import warn

import numpy as np
import pandas as pd

from .._utils import order_as_data_mapping
from ..doctools import document
from ..exceptions import PlotnineWarning
from ..mapping import aes
from .geom import geom
from .geom_path import geom_path
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    from typing import Any

    from matplotlib.axes import Axes

    from plotnine.coords.coord import coord
    from plotnine.iapi import panel_view
    from plotnine.typing import DataLike


@document
class geom_abline(geom):
    """
    Lines specified by slope and intercept

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_AES = {
        "color": "black",
        "linetype": "solid",
        "alpha": 1,
        "size": 0.5,
    }
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "inherit_aes": False,
    }
    REQUIRED_AES = {"slope", "intercept"}
    draw_legend = staticmethod(geom_path.draw_legend)

    def __init__(
        self,
        mapping: aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        data, mapping = order_as_data_mapping(data, mapping)
        slope = kwargs.pop("slope", None)
        intercept = kwargs.pop("intercept", None)

        # If nothing is set, it defaults to y=x
        if mapping is None and slope is None and intercept is None:
            slope = 1
            intercept = 0

        if slope is not None or intercept is not None:
            if mapping:
                warn(
                    "The 'intercept' and 'slope' when specified override "
                    "the aes() mapping.",
                    PlotnineWarning,
                )

            if isinstance(data, Sized) and len(data):
                warn(
                    "The 'intercept' and 'slope' when specified override "
                    "the data",
                    PlotnineWarning,
                )

            if slope is None:
                slope = 1
            if intercept is None:
                intercept = 0

            data = pd.DataFrame(
                {"intercept": np.repeat(intercept, 1), "slope": slope}
            )

            mapping = aes(intercept="intercept", slope="slope")
            kwargs["show_legend"] = False

        geom.__init__(self, mapping, data, **kwargs)

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups
        """
        ranges = coord.backtransform_range(panel_params)
        data["x"] = ranges.x[0]
        data["xend"] = ranges.x[1]
        data["y"] = ranges.x[0] * data["slope"] + data["intercept"]
        data["yend"] = ranges.x[1] * data["slope"] + data["intercept"]
        data = data.drop_duplicates()

        for _, gdata in data.groupby("group"):
            gdata.reset_index(inplace=True)
            geom_segment.draw_group(gdata, panel_params, coord, ax, **params)



from __future__ import annotations

import hashlib
from contextlib import suppress
from dataclasses import dataclass, field
from functools import cached_property
from itertools import islice
from types import SimpleNamespace as NS
from typing import TYPE_CHECKING, cast
from warnings import warn

import numpy as np
import pandas as pd

from .._utils import remove_missing
from ..exceptions import PlotnineError, PlotnineWarning
from ..mapping.aes import rename_aesthetics
from .guide import GuideElements, guide

if TYPE_CHECKING:
    from typing import Any, Optional

    from matplotlib.artist import Artist
    from matplotlib.offsetbox import PackerBase

    from plotnine.geoms.geom import geom
    from plotnine.layer import layer
    from plotnine.typing import SidePosition, TupleFloat2, TupleInt2


# See guides.py for terminology


@dataclass
class LayerParameters:
    geom: geom
    data: pd.DataFrame
    layer: layer


@dataclass
class guide_legend(guide):
    """
    Legend guide
    """

    nrow: Optional[int] = None
    """Number of rows of legends."""

    ncol: Optional[int] = None
    """Number of columns of legends."""

    byrow: bool = False
    """Whether to fill the legend row-wise or column-wise."""

    override_aes: dict[str, Any] = field(default_factory=dict)
    """Aesthetic parameters of legend key."""

    # Non-Parameter Attributes
    available_aes: set[str] = field(
        init=False, default_factory=lambda: {"any"}
    )
    """Aesthetics for which this guide can be used"""

    _layer_parameters: list[LayerParameters] = field(
        init=False, default_factory=list
    )

    def __post_init__(self):
        self._elements_cls = GuideElementsLegend
        self.elements: GuideElementsLegend

    def train(self, scale, aesthetic=None):
        """
        Create the key for the guide

        The key is a dataframe with two columns:

        - scale name : values
        - label : labels for each value

        scale name is one of the aesthetics: `x`, `y`, `color`,
        `fill`, `size`, `shape`, `alpha`, `stroke`.

        Returns this guide if training is successful and None
        if it fails
        """
        if aesthetic is None:
            aesthetic = scale.aesthetics[0]

        breaks = scale.get_bounded_breaks()
        if not breaks:
            return None

        key = pd.DataFrame(
            {aesthetic: scale.map(breaks), "label": scale.get_labels(breaks)}
        )

        if len(key) == 0:
            return None

        self.key = key

        # create a hash of the important information in the guide
        labels = " ".join(str(x) for x in self.key["label"])
        info = "\n".join(
            [
                str(self.title),
                labels,
                str(self.direction),
                self.__class__.__name__,
            ]
        )
        self.hash = hashlib.sha256(info.encode("utf-8")).hexdigest()
        return self

    def merge(self, other):
        """
        Merge overlapped guides

        For example:

        ```python
        from ggplot import *
        p = (
            ggplot(aes(x="cut", fill="cut", color="cut"), data=diamonds)
            + stat_bin()
        )
        ```

        Would create similar guides for fill and color where only
        a single guide would do
        """
        self.key = self.key.merge(other.key)
        duplicated = set(self.override_aes) & set(other.override_aes)
        if duplicated:
            msg = f"Duplicated override_aes, `{duplicated}`, are  ignored."
            warn(msg, PlotnineWarning)
        self.override_aes.update(other.override_aes)
        for ae in duplicated:
            del self.override_aes[ae]
        return self

    def create_geoms(self):
        """
        Make information needed to draw a legend for each of the layers.

        For each layer, that information is a dictionary with the geom
        to draw the guide together with the data and the parameters that
        will be used in the call to geom.
        """
        # A layer either contributes to the guide, or it does not. The
        # guide entries may be ploted in the layers
        for l in self.plot_layers:
            exclude = set()
            if isinstance(l.show_legend, dict):
                l.show_legend = rename_aesthetics(l.show_legend)
                exclude = {ae for ae, val in l.show_legend.items() if not val}
            elif l.show_legend not in (None, True):
                continue

            matched = self.legend_aesthetics(l)
            matched_set = set(matched)

            # This layer does not contribute to the legend
            if not matched_set - exclude:
                continue

            data = self.key[matched].copy()

            # Modify aesthetics

            # When doing after_scale evaluations, we only consider those
            # for the aesthetics of this legend. The reduces the spurious
            # warnings where an evaluation of another aesthetic failed yet
            # it is not needed.
            aes_modifiers = {
                ae: expr
                for ae, expr in l.mapping._scaled.items()
                if ae in matched_set
            }

            try:
                data = l.use_defaults(data, aes_modifiers)
            except PlotnineError:
                warn(
                    "Failed to apply `after_scale` modifications "
                    "to the legend.",
                    PlotnineWarning,
                )
                data = l.use_defaults(data, {})

            # override.aes in guide_legend manually changes the geom
            for ae in set(self.override_aes) & set(data.columns):
                data[ae] = self.override_aes[ae]

            data = remove_missing(
                data,
                l.geom.params["na_rm"],
                list(l.geom.REQUIRED_AES | l.geom.NON_MISSING_AES),
                f"{l.geom.__class__.__name__} legend",
            )
            self._layer_parameters.append(LayerParameters(l.geom, data, l))

        if not self._layer_parameters:
            return None
        return self

    def _calculate_rows_and_cols(
        self, elements: GuideElementsLegend
    ) -> TupleInt2:
        nrow, ncol = self.nrow, self.ncol
        nbreak = len(self.key)

        if nrow and ncol:
            if nrow * ncol < nbreak:
                raise PlotnineError(
                    "nrow x ncol needs to be larger than the number of breaks"
                )
            return nrow, ncol

        if (nrow, ncol) == (None, None):
            if elements.is_horizontal:
                nrow = int(np.ceil(nbreak / 5))
            else:
                ncol = int(np.ceil(nbreak / 20))

        if nrow is None:
            ncol = cast(int, ncol)
            nrow = int(np.ceil(nbreak / ncol))
        elif ncol is None:
            nrow = cast(int, nrow)
            ncol = int(np.ceil(nbreak / nrow))

        return nrow, ncol

    def draw(self):
        """
        Draw guide

        Returns
        -------
        out : matplotlib.offsetbox.Offsetbox
            A drawing of this legend
        """
        from matplotlib.offsetbox import HPacker, TextArea, VPacker

        from .._mpl.offsetbox import ColoredDrawingArea

        obverse = slice(0, None)
        reverse = slice(None, None, -1)
        nbreak = len(self.key)
        targets = self.theme.targets
        keys_order = reverse if self.reverse else obverse
        elements = self.elements

        # title
        title = cast(str, self.title)
        title_box = TextArea(title)
        targets.legend_title = title_box._text  # type: ignore

        # labels
        props = {"ha": elements.text.ha, "va": elements.text.va}
        labels = [TextArea(s, textprops=props) for s in self.key["label"]]
        _texts = [l._text for l in labels]  # type: ignore
        targets.legend_text_legend = _texts

        # Drawings
        drawings: list[ColoredDrawingArea] = []
        for i in range(nbreak):
            da = ColoredDrawingArea(
                elements.key_widths[i], elements.key_heights[i], 0, 0
            )

            # overlay geoms
            for params in self._layer_parameters:
                with suppress(IndexError):
                    key_data = params.data.iloc[i]
                    params.geom.draw_legend(key_data, da, params.layer)

            drawings.append(da)
        targets.legend_key = drawings

        # Match Drawings with labels to create the entries
        lookup: dict[SidePosition, tuple[type[PackerBase], slice]] = {
            "right": (HPacker, reverse),
            "left": (HPacker, obverse),
            "bottom": (VPacker, reverse),
            "top": (VPacker, obverse),
        }
        packer, slc = lookup[elements.text_position]
        if self.elements.text.is_blank:
            key_boxes = [d for d in drawings][keys_order]
        else:
            key_boxes = [
                packer(
                    children=[l, d][slc],
                    sep=elements.text.margin,
                    align=elements.text.align,
                    pad=0,
                )
                for d, l in zip(drawings, labels)
            ][keys_order]

        # Put the entries together in rows or columns
        # A chunk is either a row or a column of entries
        # for a single legend
        nrow, ncol = self._calculate_rows_and_cols(elements)
        if self.byrow:
            chunk_size = ncol
            packer_dim1, packer_dim2 = (HPacker, VPacker)
            sep1 = elements.key_spacing_x
            sep2 = elements.key_spacing_y
        else:
            chunk_size = nrow
            packer_dim1, packer_dim2 = (VPacker, HPacker)
            sep1 = elements.key_spacing_y
            sep2 = elements.key_spacing_x

        chunks = []
        for i in range(len(key_boxes)):
            start = i * chunk_size
            stop = start + chunk_size
            s = islice(key_boxes, start, stop)
            chunks.append(list(s))
            if stop >= len(key_boxes):
                break

        chunk_boxes: list[Artist] = [
            packer_dim1(children=chunk, align="left", sep=sep1, pad=0)
            for chunk in chunks
        ]

        # Put all the entries (row & columns) together
        entries_box = packer_dim2(
            children=chunk_boxes, align="baseline", sep=sep2, pad=0
        )

        # Put the title and entries together
        packer, slc = lookup[elements.title_position]

        if elements.title.is_blank:
            children: list[Artist] = [entries_box]
        else:
            children = [title_box, entries_box][slc]

        box = packer(
            children=children,
            sep=elements.title.margin,
            align=elements.title.align,
            pad=elements.margin,
        )

        return box


class GuideElementsLegend(GuideElements):
    """
    Access & calculate theming for the legend
    """

    @cached_property
    def text(self):
        size = self.theme.getp(("legend_text_legend", "size"))
        ha = self.theme.getp(("legend_text_legend", "ha"), "center")
        va = self.theme.getp(("legend_text_legend", "va"), "center")
        is_blank = self.theme.T.is_blank("legend_text_legend")

        # The original ha & va values are used by the HPacker/VPacker
        # to align the TextArea with the DrawingArea.
        # We set ha & va to values that combine best with the aligning
        # for the text area.
        align = va if self.text_position in {"left", "right"} else ha
        return NS(
            margin=self._text_margin,
            align=align,
            fontsize=size,
            ha="center",
            va="baseline",
            is_blank=is_blank,
        )

    @cached_property
    def text_position(self) -> SidePosition:
        if not (pos := self.theme.getp("legend_text_position")):
            pos = "right"
        return pos

    @cached_property
    def key_spacing_x(self) -> float:
        return self.theme.getp("legend_key_spacing_x", 0)

    @cached_property
    def key_spacing_y(self) -> float:
        return self.theme.getp("legend_key_spacing_y", 0)

    @cached_property
    def _key_dimensions(self) -> list[TupleFloat2]:
        """
        key width and key height for each legend entry

        Take a peak into data['size'] to make sure the legend key
        dimensions are big enough.
        """
        #  Note the different height sizes for the entries
        guide = cast(guide_legend, self.guide)
        min_size = (
            self.theme.getp("legend_key_width"),
            self.theme.getp("legend_key_height"),
        )

        # Find the size that fits each key in the legend,
        sizes: list[list[TupleFloat2]] = []
        for params in guide._layer_parameters:
            sizes.append([])
            get_key_size = params.geom.legend_key_size
            for i in range(len(params.data)):
                key_data = params.data.iloc[i]
                sizes[-1].append(
                    get_key_size(key_data, min_size, params.layer)
                )

        # The maximum size across each layer
        arr = np.max(sizes, axis=0)
        return [tuple(row) for row in arr]

    @cached_property
    def key_widths(self) -> list[float]:
        """
        Widths of the keys

        If legend is vertical, key widths must be equal, so we use the
        maximum. So a plot like

           (ggplot(diamonds, aes(x="cut", y="clarity"))
            + stat_sum(aes(group="cut"))
            + scale_size(range=(3, 25))
           )

        would have keys with variable heights, but fixed width.
        """
        ws = [w for w, _ in self._key_dimensions]
        if self.is_vertical:
            return [max(ws)] * len(ws)
        return ws

    @cached_property
    def key_heights(self) -> list[float]:
        """
        Heights of the keys

        If legend is horizontal, then key heights must be equal, so we
        use the maximum
        """
        hs = [h for h, _ in self._key_dimensions]
        if self.is_horizontal:
            return [max(hs)] * len(hs)
        return hs



from __future__ import annotations

import hashlib
from dataclasses import dataclass, field
from functools import cached_property
from types import SimpleNamespace as NS
from typing import TYPE_CHECKING, cast
from warnings import warn

import numpy as np
import pandas as pd
from mizani.bounds import rescale

from .._utils import get_opposite_side
from ..exceptions import PlotnineError, PlotnineWarning
from ..mapping.aes import rename_aesthetics
from ..scales.scale_continuous import scale_continuous
from .guide import GuideElements, guide

if TYPE_CHECKING:
    from typing import Literal, Optional, Sequence

    from matplotlib.artist import Artist
    from matplotlib.collections import LineCollection
    from matplotlib.offsetbox import AuxTransformBox, PackerBase
    from matplotlib.text import Text

    from plotnine import theme
    from plotnine.scales.scale import scale
    from plotnine.typing import SidePosition


@dataclass
class guide_colorbar(guide):
    """
    Guide colorbar

    Notes
    -----
    To correctly place a rasterized colorbar when saving the plot as an `svg`
    or `pdf`, you should set the `dpi` to 72 i.e. `theme(dpi=72)`{.py}.
    """

    nbin: Optional[int] = None
    """
    Number of bins for drawing a colorbar. A larger value yields
    a smoother colorbar
    """

    display: Literal["gradient", "rectangles", "raster"] = "gradient"
    """How to render the colorbar."""

    alpha: Optional[float] = None
    """
    Opacity (in the range `[0, 1]`) of the colorbar. The defualt
    `None`, is to use the opacity of the plot.
    """

    draw_ulim: bool = True
    """Whether to show the upper limit tick marks."""

    draw_llim: bool = True
    """Whether to show the lower limit tick marks. """

    # Non-Parameter Attributes
    available_aes: set[str] = field(
        init=False, default_factory=lambda: {"colour", "color", "fill"}
    )

    def __post_init__(self):
        self._elements_cls = GuideElementsColorbar
        self.elements: GuideElementsColorbar

        if self.nbin is None:
            self.nbin = 300  # if self.display == "gradient" else 300

    def train(self, scale: scale, aesthetic=None):
        self.nbin = cast(int, self.nbin)
        self.title = cast(str, self.title)

        if not isinstance(scale, scale_continuous):
            warn("colorbar guide needs continuous scales", PlotnineWarning)
            return None

        if aesthetic is None:
            aesthetic = scale.aesthetics[0]

        # Do nothing if scales are inappropriate
        if set(scale.aesthetics) & self.available_aes == 0:
            warn("colorbar guide needs appropriate scales.", PlotnineWarning)
            return None

        # value = breaks (numeric) is used for determining the
        # position of ticks
        limits = scale.limits
        breaks = scale.get_bounded_breaks()

        if not len(breaks):
            return None

        self.key = pd.DataFrame(
            {
                aesthetic: scale.map(breaks),
                "label": scale.get_labels(breaks),
                "value": breaks,
            }
        )

        bar = np.linspace(limits[0], limits[1], self.nbin)
        self.bar = pd.DataFrame({"color": scale.map(bar), "value": bar})

        labels = " ".join(str(x) for x in self.key["label"])
        info = "\n".join(
            [
                self.title,
                labels,
                " ".join(self.bar["color"].tolist()),
                self.__class__.__name__,
            ]
        )
        self.hash = hashlib.sha256(info.encode("utf-8")).hexdigest()
        return self

    def merge(self, other):
        """
        Simply discards the other guide
        """
        return self

    def create_geoms(self):
        """
        Return self if colorbar will be drawn and None if not

        This guide is not geom based
        """
        for l in self.plot_layers:
            exclude = set()
            if isinstance(l.show_legend, dict):
                l.show_legend = rename_aesthetics(l.show_legend)
                exclude = {ae for ae, val in l.show_legend.items() if not val}
            elif l.show_legend not in (None, True):
                continue

            matched = self.legend_aesthetics(l)

            # layer uses guide
            if set(matched) - exclude:
                break
        # no break, no layer uses this guide
        else:
            return None

        return self

    def draw(self):
        """
        Draw guide

        Returns
        -------
        out : matplotlib.offsetbox.Offsetbox
            A drawing of this legend
        """
        from matplotlib.offsetbox import (
            HPacker,
            TextArea,
            VPacker,
        )
        from matplotlib.transforms import IdentityTransform

        from .._mpl.offsetbox import DPICorAuxTransformBox

        self.theme = cast("theme", self.theme)

        obverse = slice(0, None)
        reverse = slice(None, None, -1)
        nbars = len(self.bar)
        elements = self.elements
        raster = self.display == "raster"

        colors = self.bar["color"].tolist()
        labels = self.key["label"].tolist()
        targets = self.theme.targets

        # .5 puts the ticks in the middle of the bars when
        # raster=False. So when raster=True the ticks are
        # in between interpolation points and the matching is
        # close though not exactly right.
        _from = self.bar["value"].min(), self.bar["value"].max()
        tick_locations = (
            rescale(self.key["value"], (0.5, nbars - 0.5), _from)
            * elements.key_height
            / nbars
        )

        # With many bins, the ticks approach the edges of the colorbar.
        # This may look odd if there is a border and the top & bottom ticks
        # partly overlap the border only because of floating point arithmetic.
        # This eliminates some of those cases so that user does no have to
        # use llim and ulim
        if nbars >= 150 and len(tick_locations) >= 2:
            tick_locations = [
                np.floor(tick_locations[0]),
                *np.round(tick_locations[1:-1]),
                np.ceil(tick_locations[-1]),
            ]

        if self.reverse:
            colors = colors[::-1]
            labels = labels[::-1]
            tick_locations = elements.key_height - tick_locations[::-1]

        auxbox = DPICorAuxTransformBox(IdentityTransform())

        # title
        title = cast(str, self.title)
        props = {"ha": elements.title.ha, "va": elements.title.va}
        title_box = TextArea(title, textprops=props)
        targets.legend_title = title_box._text  # type: ignore

        # labels
        if not self.elements.text.is_blank:
            texts = add_labels(auxbox, labels, tick_locations, elements)
            targets.legend_text_colorbar = texts

        # colorbar
        if self.display == "rectangles":
            add_segmented_colorbar(auxbox, colors, elements)
        else:
            add_gradient_colorbar(auxbox, colors, elements, raster)

        # ticks
        visible = slice(
            None if self.draw_llim else 1,
            None if self.draw_ulim else -1,
        )
        coll = add_ticks(auxbox, tick_locations[visible], elements)
        targets.legend_ticks = coll

        # frame
        frame = add_frame(auxbox, elements)
        targets.legend_frame = frame

        # title + colorbar(with labels)
        lookup: dict[SidePosition, tuple[type[PackerBase], slice]] = {
            "right": (HPacker, reverse),
            "left": (HPacker, obverse),
            "bottom": (VPacker, reverse),
            "top": (VPacker, obverse),
        }
        packer, slc = lookup[elements.title_position]

        if elements.title.is_blank:
            children: list[Artist] = [auxbox]
        else:
            children = [title_box, auxbox][slc]

        box = packer(
            children=children,
            sep=elements.title.margin,
            align=elements.title.align,
            pad=0,
        )
        return box


guide_colourbar = guide_colorbar


def add_gradient_colorbar(
    auxbox: AuxTransformBox,
    colors: Sequence[str],
    elements: GuideElementsColorbar,
    raster: bool = False,
):
    """
    Add an interpolated gradient colorbar to DrawingArea
    """
    from matplotlib.collections import QuadMesh
    from matplotlib.colors import ListedColormap

    # Special case that arises due to not so useful
    # aesthetic mapping.
    if len(colors) == 1:
        colors = [colors[0], colors[0]]

    # Number of horizontal egdes(breaks) in the grid
    # No need to create more nbreak than colors, provided
    # no. of colors = no. of breaks = no. of cmap colors
    # the shading does a perfect interpolation
    nbreak = len(colors)

    if elements.is_vertical:
        colorbar_height = elements.key_height
        colorbar_width = elements.key_width

        mesh_width = 1
        mesh_height = nbreak - 1
        linewidth = colorbar_height / mesh_height
        # Construct rectangular meshgrid
        # The values(Z) at each vertex are just the
        # normalized (onto [0, 1]) vertical distance
        x = np.array([0, colorbar_width])
        y = np.arange(0, nbreak) * linewidth
        X, Y = np.meshgrid(x, y)
        Z = Y / y.max()
    else:
        colorbar_width = elements.key_height
        colorbar_height = elements.key_width

        mesh_width = nbreak - 1
        mesh_height = 1
        linewidth = colorbar_width / mesh_width
        x = np.arange(0, nbreak) * linewidth
        y = np.array([0, colorbar_height])
        X, Y = np.meshgrid(x, y)
        Z = X / x.max()

    # As a 3D (mesh_width x mesh_height x 2) coordinates array
    coordinates = np.stack([X, Y], axis=-1)
    cmap = ListedColormap(colors)
    coll = QuadMesh(
        coordinates,
        antialiased=False,
        shading="gouraud",
        cmap=cmap,
        array=Z.ravel(),
        rasterized=raster,
    )
    auxbox.add_artist(coll)


def add_segmented_colorbar(
    auxbox: AuxTransformBox,
    colors: Sequence[str],
    elements: GuideElementsColorbar,
):
    """
    Add 'non-rastered' colorbar to AuxTransformBox
    """
    from matplotlib.collections import PolyCollection

    nbreak = len(colors)
    if elements.is_vertical:
        colorbar_height = elements.key_height
        colorbar_width = elements.key_width

        linewidth = colorbar_height / nbreak
        verts = []
        x1, x2 = 0, colorbar_width
        for i in range(nbreak):
            y1 = i * linewidth
            y2 = y1 + linewidth
            verts.append(((x1, y1), (x1, y2), (x2, y2), (x2, y1)))
    else:
        colorbar_width = elements.key_height
        colorbar_height = elements.key_width

        linewidth = colorbar_width / nbreak
        verts = []
        y1, y2 = 0, colorbar_height
        for i in range(nbreak):
            x1 = i * linewidth
            x2 = x1 + linewidth
            verts.append(((x1, y1), (x1, y2), (x2, y2), (x2, y1)))

    coll = PolyCollection(
        verts,
        facecolors=colors,
        linewidth=0,
        antialiased=False,
    )
    auxbox.add_artist(coll)


def add_ticks(auxbox, locations, elements) -> LineCollection:
    """
    Add ticks to colorbar
    """
    from matplotlib.collections import LineCollection

    segments = []
    l = elements.ticks_length
    tick_stops = np.array([0.0, l, 1 - l, 1]) * elements.key_width
    if elements.is_vertical:
        x1, x2, x3, x4 = tick_stops
        for y in locations:
            segments.extend(
                [
                    ((x1, y), (x2, y)),
                    ((x3, y), (x4, y)),
                ]
            )
    else:
        y1, y2, y3, y4 = tick_stops
        for x in locations:
            segments.extend(
                [
                    ((x, y1), (x, y2)),
                    ((x, y3), (x, y4)),
                ]
            )

    coll = LineCollection(segments)
    auxbox.add_artist(coll)
    return coll


def add_labels(
    auxbox: AuxTransformBox,
    labels: Sequence[str],
    ys: Sequence[float],
    elements: GuideElementsColorbar,
) -> list[Text]:
    """
    Return Texts added to the auxbox
    """
    from matplotlib.text import Text

    n = len(labels)
    sep = elements.text.margin
    texts: list[Text] = []
    props = {"ha": elements.text.ha, "va": elements.text.va}

    # The horizontal and vertical alignments are set in the theme
    # or dynamically calculates in GuideElements and added to the
    # themeable properties dict
    if elements.is_vertical:
        if elements.text_position == "right":
            xs = [elements.key_width + sep] * n
        else:
            xs = [-sep] * n
    else:
        xs = ys
        if elements.text_position == "bottom":
            ys = [-sep] * n
        else:
            ys = [elements.key_width + sep] * n

    for x, y, s in zip(xs, ys, labels):
        t = Text(x, y, s, **props)
        auxbox.add_artist(t)
        texts.append(t)

    return texts


def add_frame(auxbox, elements):
    """
    Add frame to colorbar
    """
    from matplotlib.patches import Rectangle
    # from .._mpl.patches import InsideStrokedRectangle as Rectangle

    width = elements.key_width
    height = elements.key_height

    if elements.is_horizontal:
        width, height = height, width

    rect = Rectangle((0, 0), width, height, facecolor="none")
    auxbox.add_artist(rect)
    return rect


class GuideElementsColorbar(GuideElements):
    """
    Access & calculate theming for the colobar
    """

    @cached_property
    def text(self):
        size = self.theme.getp(("legend_text_colorbar", "size"))
        ha = self.theme.getp(("legend_text_colorbar", "ha"))
        va = self.theme.getp(("legend_text_colorbar", "va"))
        is_blank = self.theme.T.is_blank("legend_text_colorbar")

        # Default text alignment depends on the direction of the
        # colorbar
        _loc = get_opposite_side(self.text_position)
        if self.is_vertical:
            ha = ha or _loc
            va = va or "center"
        else:
            va = va or _loc
            ha = ha or "center"

        return NS(
            margin=self._text_margin,
            align=None,
            fontsize=size,
            ha=ha,
            va=va,
            is_blank=is_blank,
        )

    @cached_property
    def text_position(self) -> SidePosition:
        if not (position := self.theme.getp("legend_text_position")):
            position = "right" if self.is_vertical else "bottom"

        if self.is_vertical and position not in ("right", "left"):
            msg = (
                "The text position for a vertical legend must be "
                "either left or right."
            )
            raise PlotnineError(msg)
        elif self.is_horizontal and position not in ("bottom", "top"):
            msg = (
                "The text position for a horizonta legend must be "
                "either top or bottom."
            )
            raise PlotnineError(msg)
        return position

    @cached_property
    def key_width(self):
        # We scale up the width only if it inherited its value
        dim = (self.is_vertical and "width") or "height"
        legend_key_dim = f"legend_key_{dim}"
        inherited = self.theme.T.get(legend_key_dim) is None
        scale = 1.45 if inherited else 1
        return np.round(self.theme.getp(legend_key_dim) * scale)

    @cached_property
    def key_height(self):
        # We scale up the height only if it inherited its value
        dim = (self.is_vertical and "height") or "width"
        legend_key_dim = f"legend_key_{dim}"
        inherited = self.theme.T.get(legend_key_dim) is None
        scale = (1.45 * 5) if inherited else 1
        return np.round(self.theme.getp(legend_key_dim) * scale)

    @cached_property
    def frame(self):
        lw = self.theme.getp(("legend_frame", "linewidth"), 0)
        return NS(linewidth=lw)

    @cached_property
    def ticks_length(self):
        return self.theme.getp("legend_ticks_length")

    @cached_property
    def ticks(self):
        lw = self.theme.getp(("legend_ticks", "linewidth"))
        return NS(linewidth=lw)



from __future__ import annotations

from typing import Optional


def get_plotnine_all(use_clipboard=True) -> Optional[str]:
    """
    Generate package level * (star) imports for plotnine

    The contents of __all__ in plotnine/__init__.py
    """
    from importlib import import_module
    from textwrap import indent

    modules = (
        "coords",
        "facets",
        "geoms",
        "ggplot",
        "guides",
        "labels",
        "mapping",
        "positions",
        "qplot",
        "scales",
        "stats",
        "themes",
        "themes.elements",
        "watermark",
    )

    comma_join = ",\n".join

    def get_all_from_module(name):
        """
        Module level imports
        """
        qname = f"plotnine.{name}"
        comment = f"# {qname}\n"
        m = import_module(qname)
        return comment + comma_join(f'"{x}"' for x in sorted(m.__all__))

    _imports = "\n".join(f"from .{name} import *" for name in modules)
    lst = indent(
        comma_join(get_all_from_module(name) for name in modules), " " * 4
    )
    _all = f"__all__ = (\n{lst},\n)"
    content = f"{_imports}\n\n{_all}"
    if use_clipboard:
        from pandas.io import clipboard

        clipboard.copy(content)  # pyright: ignore
    else:
        return content



from __future__ import annotations

import typing
from abc import ABC
from copy import copy, deepcopy
from warnings import warn

import numpy as np
from mizani.palettes import identity_pal

from .._utils.registry import Register
from ..exceptions import PlotnineError, PlotnineWarning
from ..mapping.aes import is_position_aes, rename_aesthetics
from .range import Range

if typing.TYPE_CHECKING:
    from typing import Any, Literal, Optional

    import pandas as pd
    from numpy.typing import NDArray

    from plotnine.typing import (
        ScaleBreaks,
        ScaledAestheticsName,
        ScaleLabels,
        ScaleLabelsRaw,
        ScaleLimits,
        TupleFloat2,
        TupleFloat4,
    )

    from ..iapi import range_view, scale_view


class scale(ABC, metaclass=Register):
    """
    Base class for all scales

    Parameters
    ----------
    breaks : bool | list | callable, default=True
        List of major break points. Or a callable that
        takes a tuple of limits and returns a list of breaks.
        If `True`, automatically calculate the breaks.
    expand : tuple, default=None
        Multiplicative and additive expansion constants
        that determine how the scale is expanded. If
        specified must be of length 2 or 4. Specifically the
        values are in this order:

        ```
        (mul, add)
        (mul_low, add_low, mul_high, add_high)
        ```

        For example,

        - `(0, 0)` - Do not expand.
        - `(0, 1)` - Expand lower and upper limits by 1 unit.
        - `(1, 0)` - Expand lower and upper limits by 100%.
        - `(0, 0, 0, 0)` - Do not expand, as `(0, 0)`.
        - `(0, 0, 0, 1)` - Expand upper limit by 1 unit.
        - `(0, 1, 0.1, 0)` - Expand lower limit by 1 unit and
          upper limit by 10%.
        - `(0, 0, 0.1, 2)` - Expand upper limit by 10% plus
          2 units.

        If not specified, suitable defaults are chosen.
    name : str, default=None
        Name used as the label of the scale. This is what
        shows up as the axis label or legend title. Suitable
        defaults are chosen depending on the type of scale.
    labels : bool | list | callable, default=True
        List of [](`str`). Labels at the `breaks`.
        Alternatively, a callable that takes an array_like of
        break points as input and returns a list of strings.
    limits : array_like, default=None
        Limits of the scale. Most commonly, these are the
        min & max values for the scales. For scales that
        deal with categoricals, these may be a subset or
        superset of the categories.
    na_value : scalar, default=float("nan")
        What value to assign to missing values. Default
        is to assign `np.nan`.
    palette : callable, default=None
        Function to map data points onto the scale. Most
        scales define their own palettes.
    aesthetics : list | str, default=None
        list of [](`str`). Aesthetics covered by the
        scale. These are defined by each scale and the
        user should probably not change them. Have fun.
    """

    # major breaks
    breaks = True

    # aesthetics affected by this scale
    _aesthetics: list[ScaledAestheticsName] = []

    # What to do with the NA values
    na_value: Any = np.nan

    # used as the axis label or legend title
    name: str | None = None

    # labels at the breaks
    labels: ScaleLabelsRaw = True

    # legend or any other guide
    guide: Optional[Literal["legend", "colorbar"]] = "legend"

    # (min, max) - set by user
    _limits = None

    # multiplicative and additive expansion constants
    expand: Optional[TupleFloat2 | TupleFloat4] = None

    # range class for the aesthetic
    _range_class: type[Range] = Range

    # Default palette
    _palette = identity_pal()

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                msg = "{} could not recognise parameter `{}`"
                warn(msg.format(self.__class__.__name__, k), PlotnineWarning)

        self.range = self._range_class()

        if (
            np.iterable(self.breaks)
            and np.iterable(self.labels)
            and len(self.breaks) != len(self.labels)  # type: ignore
        ):
            raise PlotnineError("Breaks and labels have unequal lengths")

        if (
            self.breaks is None
            and not is_position_aes(self.aesthetics)
            and self.guide is not None
        ):
            self.guide = None

    @property
    def aesthetics(self):
        return self._aesthetics

    @aesthetics.setter
    def aesthetics(self, value):
        if isinstance(value, str):
            value = [value]
        # TODO: Find a way to make the type checking work
        self._aesthetics = rename_aesthetics(value)  # pyright: ignore

    def __radd__(self, plot):
        """
        Add this scale to ggplot object
        """
        plot.scales.append(copy(self))
        return plot

    def palette(self, value):
        """
        Aesthetic mapping function

        Note that not all scales need to implement/provide a palette.
        For example identity & position scales do not use a palette.
        """
        return self._palette(value)  # type: ignore

    def map(self, x, limits=None):
        """
        Map every element of x

        The palette should do the real work, this should
        make sure that sensible values are sent and
        return from the palette.
        """
        raise NotImplementedError

    def train(self, x: pd.Series | NDArray):
        """
        Train scale

        Parameters
        ----------
        x :
            A column of data to train over
        """
        raise NotImplementedError

    def dimension(self, expand=None, limits=None):
        """
        Get the phyical size of the scale.
        """
        raise NotImplementedError

    def expand_limits(
        self,
        limits,  # : ScaleLimits
        expand,  # : TupleFloat2 | TupleFloat4
        coord_limits,  # : CoordRange | None
        trans,  # : Trans | Type[Trans]
    ) -> range_view:
        """
        Exand the limits of the scale
        """
        raise NotImplementedError

    def transform_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transform dataframe
        """
        raise NotImplementedError

    def transform(self, x):
        """
        Transform array|series x
        """
        raise NotImplementedError

    def inverse_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Inverse transform dataframe
        """
        raise NotImplementedError

    def inverse(self, x):
        """
        Inverse transform array|series x
        """
        raise NotImplementedError

    def view(
        self,
        limits=None,  # : Optional[ScaleLimits]
        range=None,  # : Optional[CoordRange] = None
    ) -> scale_view:
        """
        Information about the trained scale
        """
        raise NotImplementedError

    def default_expansion(
        self,
        mult: float | TupleFloat2 = 0,
        add: float | TupleFloat2 = 0,
        expand=True,
    ) -> TupleFloat4:
        """
        Get default expansion for this scale
        """
        if not expand:
            return (0, 0, 0, 0)

        if self.expand:
            if len(self.expand) == 2:
                mult, add = self.expand
            else:
                # n == 4:
                mult = self.expand[0], self.expand[2]
                add = self.expand[1], self.expand[3]

        if isinstance(mult, (float, int)):
            mult = (mult, mult)

        if isinstance(add, (float, int)):
            add = (add, add)

        if len(mult) != 2:
            raise ValueError(
                "The scale expansion multiplication factor should "
                "either be a single float, or a tuple of two floats. "
            )

        if len(add) != 2:
            raise ValueError(
                "The scale expansion addition constant should "
                "either be a single float, or a tuple of two floats. "
            )

        return (mult[0], add[0], mult[1], add[1])

    def clone(self):
        return deepcopy(self)

    def reset(self):
        """
        Set the range of the scale to None.

        i.e Forget all the training
        """
        self.range.reset()

    def is_empty(self) -> bool:
        """
        Whether the scale has size information
        """
        if not hasattr(self, "range"):
            return True
        return self.range.is_empty() and self._limits is None

    @property
    def limits(self) -> ScaleLimits:
        raise NotImplementedError

    @limits.setter
    def limits(
        self,
        value,  # : ScaleLimitsRaw
    ):
        raise NotImplementedError

    def train_df(self, df: pd.DataFrame):
        """
        Train scale from a dataframe
        """
        aesthetics = sorted(set(self.aesthetics) & set(df.columns))
        for ae in aesthetics:
            self.train(df[ae])

    def map_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Map df
        """
        if len(df) == 0:
            return df

        aesthetics = set(self.aesthetics) & set(df.columns)
        for ae in aesthetics:
            df[ae] = self.map(df[ae])

        return df

    def get_labels(
        self,
        breaks=None,  # : Optional[ScaleBreaks]
    ) -> ScaleLabels:
        """
        Get labels, calculating them if required
        """
        raise NotImplementedError

    def get_breaks(
        self,
        limits=None,  # : Optional[ScaleLimits]
    ) -> ScaleBreaks:
        """
        Get Breaks
        """
        raise NotImplementedError

    def get_bounded_breaks(
        self,
        limits=None,  # : Optional[ScaleLimits]
    ) -> ScaleBreaks:
        """
        Return Breaks that are within the limits
        """
        raise NotImplementedError



import numpy as np
import pandas as pd

from .._utils import resolution
from ..doctools import document
from .stat import stat


@document
class stat_boxplot(stat):
    """
    Compute boxplot statistics

    {usage}

    Parameters
    ----------
    {common_parameters}
    coef : float, default=1.5
        Length of the whiskers as a multiple of the Interquartile
        Range.

    See Also
    --------
    plotnine.geom_boxplot
    """

    _aesthetics_doc = """
    {aesthetics_table}

    **Options for computed aesthetics**

    ```python
    "width"  # width of boxplot
    "lower"  # lower hinge, 25% quantile
    "middle" # median, 50% quantile
    "upper"  # upper hinge, 75% quantile

    # lower edge of notch, computed as;
    # median - 1.58 * IQR / sqrt(n)
    "notchlower"

    # upper edge of notch, computed as;
    # median + 1.58 * IQR / sqrt(n)
    "notchupper"

    # lower whisker, computed as; smallest observation
    # greater than or equal to lower hinge - 1.5 * IQR
    "ymin"

    # upper whisker, computed as; largest observation
    # less than or equal to upper hinge + 1.5 * IQR
    "ymax"
    ```

        'n'     # Number of observations at a position

    Calculated aesthetics are accessed using the `after_stat` function.
    e.g. `after_stat('width')`{.py}.
    """

    REQUIRED_AES = {"x", "y"}
    NON_MISSING_AES = {"weight"}
    DEFAULT_PARAMS = {
        "geom": "boxplot",
        "position": "dodge",
        "na_rm": False,
        "coef": 1.5,
        "width": None,
    }
    CREATES = {
        "lower",
        "upper",
        "middle",
        "ymin",
        "ymax",
        "outliers",
        "notchupper",
        "notchlower",
        "width",
        "relvarwidth",
        "n",
    }

    def setup_data(self, data):
        if "x" not in data:
            data["x"] = 0
        return data

    def setup_params(self, data):
        if self.params["width"] is None:
            x = data.get("x", 0)
            self.params["width"] = resolution(x, False) * 0.75
        return self.params

    @classmethod
    def compute_group(cls, data, scales, **params):
        n = len(data)
        y = data["y"].to_numpy()
        if "weight" in data:
            weights = data["weight"]
            total_weight = np.sum(weights)
        else:
            weights = None
            total_weight = len(y)
        res = weighted_boxplot_stats(y, weights=weights, whis=params["coef"])

        if len(np.unique(data["x"])) > 1:
            width = np.ptp(data["x"]) * 0.9
        else:
            width = params["width"]

        if isinstance(data["x"].dtype, pd.CategoricalDtype):
            x = data["x"].iloc[0]
        else:
            x = np.mean([data["x"].min(), data["x"].max()])

        d = {
            "ymin": res["whislo"],
            "lower": res["q1"],
            "middle": [res["med"]],
            "upper": res["q3"],
            "ymax": res["whishi"],
            "outliers": [res["fliers"]],
            "notchupper": res["cihi"],
            "notchlower": res["cilo"],
            "x": x,
            "width": width,
            "relvarwidth": np.sqrt(total_weight),
            "n": n,
        }
        return pd.DataFrame(d)


def weighted_percentile(a, q, weights=None):
    """
    Compute the weighted q-th percentile of data

    Parameters
    ----------
    a : array_like
        Input that can be converted into an array.
    q : array_like[float]
        Percentile or sequence of percentiles to compute. Must be int
        the range [0, 100]
    weights : array_like
        Weights associated with the input values.
    """
    # Calculate and interpolate weighted percentiles
    # method derived from https://en.wikipedia.org/wiki/Percentile
    # using numpy's standard C = 1
    if weights is None:
        weights = np.ones(len(a))

    weights = np.asarray(weights)
    q = np.asarray(q)

    C = 1
    idx_s = np.argsort(a)
    a_s = a[idx_s]
    w_n = weights[idx_s]
    S_N = np.sum(weights)
    S_n = np.cumsum(w_n)
    p_n = (S_n - C * w_n) / (S_N + (1 - 2 * C) * w_n)
    pcts = np.interp(q / 100.0, p_n, a_s)
    return pcts


def weighted_boxplot_stats(x, weights=None, whis=1.5):
    """
    Calculate weighted boxplot plot statistics

    Parameters
    ----------
    x : array_like
        Data
    weights : array_like
        Weights associated with the data.
    whis : float
        Position of the whiskers beyond the interquartile range.
        The data beyond the whisker are considered outliers.

        If a float, the lower whisker is at the lowest datum above
        `Q1 - whis*(Q3-Q1)`, and the upper whisker at the highest
        datum below `Q3 + whis*(Q3-Q1)`, where Q1 and Q3 are the
        first and third quartiles.  The default value of
        `whis = 1.5` corresponds to Tukey's original definition of
        boxplots.

    Notes
    -----
    This method adapted from Matplotlibs boxplot_stats. The key difference
    is the use of a weighted percentile calculation and then using linear
    interpolation to map weight percentiles back to data.
    """
    if weights is None:
        q1, med, q3 = np.percentile(x, (25, 50, 75))
        n = len(x)
    else:
        q1, med, q3 = weighted_percentile(x, (25, 50, 75), weights)
        n = np.sum(weights)

    iqr = q3 - q1
    mean = np.average(x, weights=weights)
    cilo = med - 1.58 * iqr / np.sqrt(n)
    cihi = med + 1.58 * iqr / np.sqrt(n)

    # low extreme
    loval = q1 - whis * iqr
    lox = x[x >= loval]
    whislo = q1 if (len(lox) == 0 or np.min(lox) > q1) else np.min(lox)

    # high extreme
    hival = q3 + whis * iqr
    hix = x[x <= hival]
    whishi = q3 if (len(hix) == 0 or np.max(hix) < q3) else np.max(hix)

    bpstats = {
        "fliers": x[(x < whislo) | (x > whishi)],
        "mean": mean,
        "med": med,
        "q1": q1,
        "q3": q3,
        "iqr": iqr,
        "whislo": whislo,
        "whishi": whishi,
        "cilo": cilo,
        "cihi": cihi,
    }
    return bpstats



from __future__ import annotations

import typing
import warnings

import numpy as np
import pandas as pd

from .._utils import log
from ..coords import coord_flip
from ..exceptions import PlotnineWarning
from ..scales.scale_continuous import scale_continuous as ScaleContinuous
from .annotate import annotate
from .geom_path import geom_path
from .geom_rug import geom_rug

if typing.TYPE_CHECKING:
    from typing import Any, Literal, Optional, Sequence

    from matplotlib.axes import Axes
    from mizani.transforms import trans

    from plotnine.coords.coord import coord
    from plotnine.facets.layout import Layout
    from plotnine.geoms.geom import geom
    from plotnine.iapi import panel_view
    from plotnine.typing import (
        AnyArray,
        TupleFloat2,
        TupleFloat3,
        TupleFloat4,
    )


class _geom_logticks(geom_rug):
    """
    Internal geom implementing drawing of annotation_logticks
    """

    DEFAULT_AES = {}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "sides": "bl",
        "alpha": 1,
        "color": "black",
        "size": 0.5,
        "linetype": "solid",
        "lengths": (0.036, 0.0225, 0.012),
        "base": 10,
    }
    draw_legend = staticmethod(geom_path.draw_legend)

    def draw_layer(
        self, data: pd.DataFrame, layout: Layout, coord: coord, **params: Any
    ):
        """
        Draw ticks on every panel
        """
        for pid in layout.layout["PANEL"]:
            ploc = pid - 1
            panel_params = layout.panel_params[ploc]
            ax = layout.axs[ploc]
            self.draw_panel(data, panel_params, coord, ax, **params)

    @staticmethod
    def _check_log_scale(
        base: Optional[float],
        sides: str,
        panel_params: panel_view,
        coord: coord,
    ) -> TupleFloat2:
        """
        Check the log transforms

        Parameters
        ----------
        base : float | None
            Base of the logarithm in which the ticks will be
            calculated. If `None`, the base of the log transform
            the scale will be used.
        sides : str, default="bl"
            Sides onto which to draw the marks. Any combination
            chosen from the characters `btlr`, for *bottom*, *top*,
            *left* or *right* side marks. If `coord_flip()` is used,
            these are the sides *before* the flip.
        panel_params : panel_view
            `x` and `y` view scale values.
        coord : coord
            Coordinate (e.g. coord_cartesian) system of the geom.

        Returns
        -------
        out : tuple
            The bases (base_x, base_y) to use when generating the ticks.
        """

        def is_log_trans(t: trans) -> bool:
            return hasattr(t, "base") and t.__class__.__name__.startswith(
                "log"
            )

        def get_base(sc, ubase: Optional[float]) -> float:
            ae = sc.aesthetics[0]

            if not isinstance(sc, ScaleContinuous) or not is_log_trans(
                sc.trans
            ):
                warnings.warn(
                    f"annotation_logticks for {ae}-axis which does not have "
                    "a log scale. The logticks may not make sense.",
                    PlotnineWarning,
                )
                return 10 if ubase is None else ubase

            base = sc.trans.base  # pyright: ignore
            if ubase is not None and base != ubase:
                warnings.warn(
                    f"The x-axis is log transformed in base={base} ,"
                    "but the annotation_logticks are computed in base="
                    f"{ubase}",
                    PlotnineWarning,
                )
                return ubase
            return base

        base_x, base_y = 10, 10
        x_scale = panel_params.x.scale
        y_scale = panel_params.y.scale

        if isinstance(coord, coord_flip):
            x_scale, y_scale = y_scale, x_scale
            base_x, base_y = base_y, base_x

        if "t" in sides or "b" in sides:
            base_x = get_base(x_scale, base)

        if "l" in sides or "r" in sides:
            base_y = get_base(y_scale, base)

        return base_x, base_y

    @staticmethod
    def _calc_ticks(
        value_range: TupleFloat2, base: float
    ) -> tuple[AnyArray, AnyArray, AnyArray]:
        """
        Calculate tick marks within a range

        Parameters
        ----------
        value_range: tuple
            Range for which to calculate ticks.

        base : number
            Base of logarithm

        Returns
        -------
        out: tuple
            (major, middle, minor) tick locations
        """

        def _minor(x: Sequence[Any], mid_idx: int) -> AnyArray:
            return np.hstack([x[1:mid_idx], x[mid_idx + 1 : -1]])

        # * Calculate the low and high powers,
        # * Generate for all intervals in along the low-high power range
        #   The intervals are in normal space
        # * Calculate evenly spaced breaks in normal space, then convert
        #   them to log space.
        low = np.floor(value_range[0])
        high = np.ceil(value_range[1])
        arr = base ** np.arange(low, float(high + 1))
        n_ticks = int(np.round(base) - 1)
        breaks = [
            log(np.linspace(b1, b2, n_ticks + 1), base)
            for (b1, b2) in list(zip(arr, arr[1:]))
        ]

        # Partition the breaks in the 3 groups
        major = np.array([x[0] for x in breaks] + [breaks[-1][-1]])
        if n_ticks % 2:
            mid_idx = n_ticks // 2
            middle = np.array([x[mid_idx] for x in breaks])
            minor = np.hstack([_minor(x, mid_idx) for x in breaks])
        else:
            middle = np.array([])
            minor = np.hstack([x[1:-1] for x in breaks])

        return major, middle, minor

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: coord,
        ax: Axes,
        **params: Any,
    ):
        # Any passed data is ignored, the relevant data is created
        sides = params["sides"]
        lengths = params["lengths"]
        _aesthetics = {
            "size": params["size"],
            "color": params["color"],
            "alpha": params["alpha"],
            "linetype": params["linetype"],
        }

        def _draw(
            geom: geom,
            axis: Literal["x", "y"],
            tick_positions: tuple[AnyArray, AnyArray, AnyArray],
        ):
            for position, length in zip(tick_positions, lengths):
                data = pd.DataFrame({axis: position, **_aesthetics})
                geom.draw_group(
                    data, panel_params, coord, ax, length=length, **params
                )

        if isinstance(coord, coord_flip):
            tick_range_x = panel_params.y.range
            tick_range_y = panel_params.x.range
        else:
            tick_range_x = panel_params.x.range
            tick_range_y = panel_params.y.range

        # these are already flipped iff coord_flip
        base_x, base_y = self._check_log_scale(
            params["base"], sides, panel_params, coord
        )

        if "b" in sides or "t" in sides:
            tick_positions = self._calc_ticks(tick_range_x, base_x)
            _draw(self, "x", tick_positions)

        if "l" in sides or "r" in sides:
            tick_positions = self._calc_ticks(tick_range_y, base_y)
            _draw(self, "y", tick_positions)


class annotation_logticks(annotate):
    """
    Marginal log ticks.

    If added to a plot that does not have a log10 axis
    on the respective side, a warning will be issued.

    Parameters
    ----------
    sides :
        Sides onto which to draw the marks. Any combination
        chosen from the characters `btlr`, for *bottom*, *top*,
        *left* or *right* side marks. If `coord_flip()` is used,
        these are the sides *after* the flip.
    alpha :
        Transparency of the ticks
    color :
        Colour of the ticks
    size :
        Thickness of the ticks
    linetype :
        Type of line
    lengths:
        length of the ticks drawn for full / half / tenth
        ticks relative to panel size
    base :
        Base of the logarithm in which the ticks will be
        calculated. If `None`, the base used to log transform
        the scale will be used.
    """

    def __init__(
        self,
        sides: str = "bl",
        alpha: float = 1,
        color: str | TupleFloat3 | TupleFloat4 = "black",
        size: float = 0.5,
        linetype: Literal["solid", "dashed", "dashdot", "dotted"]
        | Sequence[float] = "solid",
        lengths: TupleFloat3 = (0.036, 0.0225, 0.012),
        base: float | None = None,
    ):
        if len(lengths) != 3:
            raise ValueError(
                "length for annotation_logticks must be a tuple of 3 floats"
            )

        self._annotation_geom = _geom_logticks(
            sides=sides,
            alpha=alpha,
            color=color,
            size=size,
            linetype=linetype,
            lengths=lengths,
            base=base,
        )



from __future__ import annotations

import typing
from copy import copy, deepcopy
from functools import cached_property
from typing import overload

from ..exceptions import PlotnineError
from ..options import get_option, set_option
from .targets import ThemeTargets
from .themeable import Themeables, themeable

if typing.TYPE_CHECKING:
    from typing import Type

    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
    from typing_extensions import Self

    from plotnine import ggplot


# All complete themes are initiated with these rcparams. They
# can be overridden.
default_rcparams = {
    "axes.axisbelow": "True",
    "font.sans-serif": [
        "Helvetica",
        "DejaVu Sans",  # MPL ships with this one
        "Avant Garde",
        "Computer Modern Sans serif",
        "Arial",
    ],
    "font.serif": [
        "Times",
        "Palatino",
        "New Century Schoolbook",
        "Bookman",
        "Computer Modern Roman",
        "Times New Roman",
    ],
    "lines.antialiased": "True",
    "patch.antialiased": "True",
    "timezone": "UTC",
}


class theme:
    """
    Base class for themes

    In general, only complete themes should subclass this class.

    Parameters
    ----------
    complete : bool
        Themes that are complete will override any existing themes.
        themes that are not complete (ie. partial) will add to or
        override specific elements of the current theme. e.g:

        ```python
        theme_gray() + theme_xkcd()
        ```

        will be completely determined by [](`~plotnine.themes.theme_xkcd`),
        but:

        ```python
        theme_gray() + theme(axis_text_x=element_text(angle=45))
        ```

        will only modify the x-axis text.
    kwargs: dict
        kwargs are `themeables`. The themeables are elements that are
        subclasses of `themeable`. Many themeables are defined using
        theme elements i.e

        - [](`~plotnine.themes.element_line`)
        - [](`~plotnine.themes.element_rect`)
        - [](`~plotnine.themes.element_text`)

        These simply bind together all the aspects of a themeable
        that can be themed. See [](`~plotnine.themes.themeable.themeable`).

    Notes
    -----
    When subclassing, make sure to call `theme.__init__`{.py}.
    After which you can customise `self._rcParams`{.py} within
    the `__init__` method of the new theme. The `rcParams`
    should not be modified after that.
    """

    complete: bool

    # This is set when the figure is created,
    # it is useful at legend drawing time and
    # when applying the theme.
    plot: ggplot
    figure: Figure
    axs: list[Axes]

    # Dictionary to collect matplotlib objects that will
    # be targeted for theming by the themeables
    # It is initialised in the setup method.
    targets: ThemeTargets

    def __init__(
        self,
        complete=False,
        # Generate themeables keyword parameters with
        #
        #     from plotnine.themes.themeable import themeable
        #     for name in themeable.registry():
        #         print(f'{name}=None,')
        axis_title_x=None,
        axis_title_y=None,
        axis_title=None,
        legend_title=None,
        legend_text_legend=None,
        legend_text_colorbar=None,
        legend_text=None,
        plot_title=None,
        plot_subtitle=None,
        plot_caption=None,
        strip_text_x=None,
        strip_text_y=None,
        strip_text=None,
        title=None,
        axis_text_x=None,
        axis_text_y=None,
        axis_text=None,
        text=None,
        axis_line_x=None,
        axis_line_y=None,
        axis_line=None,
        axis_ticks_minor_x=None,
        axis_ticks_minor_y=None,
        axis_ticks_major_x=None,
        axis_ticks_major_y=None,
        axis_ticks_major=None,
        axis_ticks_minor=None,
        axis_ticks_x=None,
        axis_ticks_y=None,
        axis_ticks=None,
        legend_ticks=None,
        panel_grid_major_x=None,
        panel_grid_major_y=None,
        panel_grid_minor_x=None,
        panel_grid_minor_y=None,
        panel_grid_major=None,
        panel_grid_minor=None,
        panel_grid=None,
        line=None,
        legend_key=None,
        legend_frame=None,
        legend_background=None,
        legend_box_background=None,
        panel_background=None,
        panel_border=None,
        plot_background=None,
        strip_background_x=None,
        strip_background_y=None,
        strip_background=None,
        rect=None,
        axis_ticks_length_major_x=None,
        axis_ticks_length_major_y=None,
        axis_ticks_length_major=None,
        axis_ticks_length_minor_x=None,
        axis_ticks_length_minor_y=None,
        axis_ticks_length_minor=None,
        axis_ticks_length=None,
        axis_ticks_pad_major_x=None,
        axis_ticks_pad_major_y=None,
        axis_ticks_pad_major=None,
        axis_ticks_pad_minor_x=None,
        axis_ticks_pad_minor_y=None,
        axis_ticks_pad_minor=None,
        axis_ticks_pad=None,
        axis_ticks_direction_x=None,
        axis_ticks_direction_y=None,
        axis_ticks_direction=None,
        panel_spacing_x=None,
        panel_spacing_y=None,
        panel_spacing=None,
        plot_margin_left=None,
        plot_margin_right=None,
        plot_margin_top=None,
        plot_margin_bottom=None,
        plot_margin=None,
        panel_ontop=None,
        aspect_ratio=None,
        dpi=None,
        figure_size=None,
        legend_box=None,
        legend_box_margin=None,
        legend_box_just=None,
        legend_justification_right=None,
        legend_justification_left=None,
        legend_justification_top=None,
        legend_justification_bottom=None,
        legend_justification_inside=None,
        legend_justification=None,
        legend_direction=None,
        legend_key_width=None,
        legend_key_height=None,
        legend_key_size=None,
        legend_ticks_length=None,
        legend_margin=None,
        legend_box_spacing=None,
        legend_spacing=None,
        legend_position_inside=None,
        legend_position=None,
        legend_title_position=None,
        legend_text_position=None,
        legend_key_spacing_x=None,
        legend_key_spacing_y=None,
        legend_key_spacing=None,
        strip_align_x=None,
        strip_align_y=None,
        strip_align=None,
        svg_usefonts=None,
        **kwargs,
    ):
        self.themeables = Themeables()
        self.complete = complete

        if complete:
            self._rcParams = deepcopy(default_rcparams)
        else:
            self._rcParams = {}

        # Themeables
        official_themeables = themeable.registry()
        locals_args = dict(locals())
        it = (
            (name, element)
            for name, element in locals_args.items()
            if element is not None and name in official_themeables
        )
        new = themeable.from_class_name

        for name, element in it:
            self.themeables[name] = new(name, element)

        # Unofficial themeables for extensions
        # or those that have been deprecated
        for name, element in kwargs.items():
            self.themeables[name] = new(name, element)

    def __eq__(self, other: object) -> bool:
        """
        Test if themes are equal

        Mostly for testing purposes
        """
        return other is self or (
            isinstance(other, type(self))
            and other.themeables == self.themeables
            and other.rcParams == self.rcParams
        )

    @cached_property
    def T(self):
        """
        Convenient access to the themeables
        """
        return self.themeables

    @cached_property
    def getp(self):
        """
        Convenient access into the properties of the themeables
        """
        return self.themeables.getp

    def apply(self):
        """
        Apply this theme, then apply additional modifications in order.

        This method will be called once after plot has completed.
        Subclasses that override this method should make sure that the
        base class method is called.
        """
        for th in self.T.values():
            th.apply(self)

    def setup(self, plot: ggplot):
        """
        Setup theme for applying

        This method will be called when the figure and axes have been created
        but before any plotting or other artists have been added to the
        figure. This method gives the theme and the elements references to
        the figure and/or axes.

        It also initialises where the artists to be themed will be stored.
        """
        self.plot = plot
        self.figure = plot.figure
        self.axs = plot.axs
        self.targets = ThemeTargets()
        self._add_default_themeable_properties()
        self.T.setup(self)

    def _add_default_themeable_properties(self):
        """
        Add default themeable properties that depend depend on the plot

        Some properties may be left unset (None) and their final values are
        best worked out dynamically after the plot has been built, but
        before the themeables are applied.

        This is where the theme is modified to add those values.
        """
        self._smart_title_and_subtitle_ha()

    @property
    def rcParams(self):
        """
        Return rcParams dict for this theme.

        Notes
        -----
        Subclasses should not need to override this method method as long as
        self._rcParams is constructed properly.

        rcParams are used during plotting. Sometimes the same theme can be
        achieved by setting rcParams before plotting or a apply
        after plotting. The choice of how to implement it is is a matter of
        convenience in that case.

        There are certain things can only be themed after plotting. There
        may not be an rcParam to control the theme or the act of plotting
        may cause an entity to come into existence before it can be themed.

        """
        try:
            rcParams = deepcopy(self._rcParams)
        except NotImplementedError:
            # deepcopy raises an error for objects that are drived from or
            # composed of matplotlib.transform.TransformNode.
            # Not desirable, but probably requires upstream fix.
            # In particular, XKCD uses matplotlib.patheffects.withStrok
            rcParams = copy(self._rcParams)

        for th in self.T.values():
            rcParams.update(th.rcParams)
        return rcParams

    def add_theme(self, other: theme) -> theme:
        """
        Add themes together

        Subclasses should not override this method.

        This will be called when adding two instances of class 'theme'
        together.
        A complete theme will annihilate any previous themes. Partial themes
        can be added together and can be added to a complete theme.
        """
        if other.complete:
            return other

        self.themeables.update(deepcopy(other.themeables))
        return self

    def __add__(self, other: theme) -> theme:
        """
        Add other theme to this theme
        """
        if not isinstance(other, theme):
            msg = f"Adding theme failed. {other} is not a theme"
            raise PlotnineError(msg)
        self = deepcopy(self)
        return self.add_theme(other)

    @overload
    def __radd__(self, other: theme) -> theme: ...

    @overload
    def __radd__(self, other: ggplot) -> ggplot: ...

    def __radd__(self, other: theme | ggplot) -> theme | ggplot:
        """
        Add theme to ggplot object or to another theme

        This will be called in one of two ways:

        ```python
        ggplot() + theme()
        theme1() + theme2()
        ```

        In both cases, `self` is the [](`~plotnine.themes.theme`)
        on the right hand side.

        Subclasses should not override this method.
        """
        # ggplot() + theme, get theme
        # if hasattr(other, 'theme'):
        if not isinstance(other, theme):
            if self.complete:
                other.theme = self
            else:
                # If no theme has been added yet,
                # we modify the default theme
                other.theme = other.theme or theme_get()
                other.theme = other.theme.add_theme(self)
            return other
        # theme1 + theme2
        else:
            if self.complete:
                # e.g. other + theme_gray()
                return self
            else:
                # e.g. other + theme(...)
                return other.add_theme(self)

    def __iadd__(self, other: theme) -> Self:
        """
        Add theme to theme
        """
        self.add_theme(other)
        return self

    def __deepcopy__(self, memo: dict) -> theme:
        """
        Deep copy without copying the figure
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        shallow = {"plot", "figure", "axs"}
        skip = {"targets"}
        for key, item in old.items():
            if key in skip:
                continue
            elif key in shallow:
                new[key] = item
                memo[id(new[key])] = new[key]
            else:
                new[key] = deepcopy(item, memo)

        return result

    def to_retina(self) -> theme:
        """
        Return a retina-sized version of this theme

        The result is a theme that has double the dpi.
        """
        dpi = self.getp("dpi")
        return self + theme(dpi=dpi * 2)

    def _smart_title_and_subtitle_ha(self):
        """
        Smartly add the horizontal alignment for the title and subtitle
        """
        from .elements import element_text

        has_title = bool(
            self.plot.labels.get("title", "")
        ) and not self.T.is_blank("plot_title")
        has_subtitle = bool(
            self.plot.labels.get("subtitle", "")
        ) and not self.T.is_blank("plot_subtitle")

        title_ha = self.getp(("plot_title", "ha"))
        subtitle_ha = self.getp(("plot_subtitle", "ha"))

        default_title_ha, default_subtitle_ha = "center", "left"
        kwargs = {}

        if has_title and title_ha is None:
            if has_subtitle and not subtitle_ha:
                title_ha = default_subtitle_ha
            else:
                title_ha = default_title_ha
            kwargs["plot_title"] = element_text(ha=title_ha)

        if has_subtitle and subtitle_ha is None:
            subtitle_ha = default_subtitle_ha
            kwargs["plot_subtitle"] = element_text(ha=subtitle_ha)

        if kwargs:
            self += theme(**kwargs)


def theme_get() -> theme:
    """
    Return the default theme

    The default theme is the one set (using [](`~plotnine.themes.theme_set`))
    by the user. If none has been set, then [](`~plotnine.themes.theme_gray`)
    is the default.
    """
    from .theme_gray import theme_gray

    _theme = get_option("current_theme")
    if isinstance(_theme, type):
        _theme = _theme()
    return _theme or theme_gray()


def theme_set(new: theme | Type[theme]) -> theme:
    """
    Change the current(default) theme

    Parameters
    ----------
    new : theme
        New default theme

    Returns
    -------
    out : theme
        Previous theme
    """
    if not isinstance(new, theme) and not issubclass(new, theme):
        raise PlotnineError("Expecting object to be a theme")

    out: theme = get_option("current_theme")
    set_option("current_theme", new)
    return out


def theme_update(**kwargs: themeable):
    """
    Modify elements of the current theme

    Parameters
    ----------
    kwargs : dict
        Theme elements
    """
    assert "complete" not in kwargs
    theme_set(theme_get() + theme(**kwargs))  # pyright: ignore



"""
Theme elements used to decorate the graph.
"""

from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING

from .element_base import element_base
from .margin import Margin

if TYPE_CHECKING:
    from typing import Any, Literal, Optional, Sequence

    from plotnine import theme
    from plotnine.typing import TupleFloat3, TupleFloat4


class element_text(element_base):
    """
    Theme element: Text

    Parameters
    ----------
    family :
        Font family. See [](`~matplotlib.text.Text.set_family`)
        for supported values.
    style :
        Font style
    color :
        Text color
    weight :
        Should be one of `normal`, `bold`, `heavy`, `light`,
        `ultrabold` or `ultralight`.
    size :
        text size
    ha :
        Horizontal Alignment.
    va :
        Vertical alignment.
    rotation :
        Rotation angle in the range [0, 360]
    linespacing : float
        Line spacing
    backgroundcolor :
        Background color
    margin :
        Margin around the text. The keys are
        `t`, `b`, `l`, `r` and `units`.
        The `tblr` keys are floats.
        The `units` is one of `pt`, `lines` or `in`.
        Not all text themeables support margin parameters and other
        than the `units`, only some of the other keys may apply.
    kwargs :
        Parameters recognised by [](`~matplotlib.text.Text`)

    Notes
    -----
    [](`~plotnine.themes.element_text`) will accept parameters that
    conform to the **ggplot2** *element_text* API, but it is preferable
    the **Matplotlib** based API described above.
    """

    def __init__(
        self,
        family: Optional[str | list[str]] = None,
        style: Optional[str | Sequence[str]] = None,
        weight: Optional[int | str | Sequence[int | str]] = None,
        color: Optional[
            str
            | TupleFloat3
            | TupleFloat4
            | Sequence[str | TupleFloat3 | TupleFloat4]
        ] = None,
        size: Optional[float | Sequence[float]] = None,
        ha: Optional[Literal["center", "left", "right"] | float] = None,
        va: Optional[
            Literal["center", "top", "bottom", "baseline", "center_baseline"]
            | float
        ] = None,
        rotation: Optional[
            Literal["vertical", "horizontal"]
            | float
            | Sequence[Literal["vertical", "horizontal"]]
            | Sequence[float]
        ] = None,
        linespacing: Optional[float] = None,
        backgroundcolor: Optional[
            str
            | TupleFloat3
            | TupleFloat4
            | Sequence[str | TupleFloat3 | TupleFloat4]
        ] = None,
        margin: Optional[
            dict[Literal["t", "b", "l", "r", "units"], Any]
        ] = None,
        **kwargs: Any,
    ):
        # ggplot2 translation
        with suppress(KeyError):
            linespacing = kwargs.pop("lineheight")
        with suppress(KeyError):
            color = color or kwargs.pop("colour")
        with suppress(KeyError):
            _face = kwargs.pop("face")
            if _face == "plain":
                style = "normal"
            elif _face == "italic":
                style = "italic"
            elif _face == "bold":
                weight = "bold"
            elif _face == "bold.italic":
                style = "italic"
                weight = "bold"
        with suppress(KeyError):
            ha = self._translate_hjust(kwargs.pop("hjust"))
        with suppress(KeyError):
            va = self._translate_vjust(kwargs.pop("vjust"))
        with suppress(KeyError):
            rotation = kwargs.pop("angle")

        super().__init__()
        self.properties.update(**kwargs)
        if margin is not None:
            self.properties["margin"] = Margin(self, **margin)

        # Use the parameters that have been set
        names = (
            "backgroundcolor",
            "color",
            "family",
            "ha",
            "linespacing",
            "rotation",
            "size",
            "style",
            "va",
            "weight",
        )
        variables = locals()
        for name in names:
            if variables[name] is not None:
                self.properties[name] = variables[name]

    def setup(self, theme: theme, themeable_name: str):
        """
        Setup the theme_element before drawing
        """
        if m := self.properties.get("margin"):
            m.theme = theme
            m.themeable_name = themeable_name

    def _translate_hjust(
        self, just: float
    ) -> Literal["left", "right", "center"]:
        """
        Translate ggplot2 justification from [0, 1] to left, right, center.
        """
        if just == 0:
            return "left"
        elif just == 1:
            return "right"
        else:
            return "center"

    def _translate_vjust(
        self, just: float
    ) -> Literal["top", "bottom", "center"]:
        """
        Translate ggplot2 justification from [0, 1] to top, bottom, center.
        """
        if just == 0:
            return "bottom"
        elif just == 1:
            return "top"
        else:
            return "center"



from __future__ import annotations

import re
import typing
from collections.abc import Iterable, Sequence
from contextlib import suppress
from copy import deepcopy
from dataclasses import fields
from typing import Any, Dict

import pandas as pd

from ..iapi import labels_view
from .evaluation import after_stat, stage

if typing.TYPE_CHECKING:
    from typing import Protocol, TypeVar

    class ColorOrColour(Protocol):
        color: Any
        colour: Any

    THasAesNames = TypeVar(
        "THasAesNames",
        bound=Sequence[str] | dict[str, Any] | ColorOrColour,
    )

__all__ = ("aes",)

X_AESTHETICS = {"x", "xmin", "xmax", "xend", "xintercept"}
Y_AESTHETICS = {"y", "ymin", "ymax", "yend", "yintercept"}

ALL_AESTHETICS = {
    "alpha",
    "angle",
    "color",
    "colour",
    "fill",
    "group",
    "intercept",
    "label",
    "lineheight",
    "linetype",
    "lower",
    "middle",
    "radius",
    "sample",
    "shape",
    "size",
    "slope",
    "stroke",
    "upper",
    "weight",
    *X_AESTHETICS,
    *Y_AESTHETICS,
}

POSITION_AESTHETICS = X_AESTHETICS | Y_AESTHETICS

SCALED_AESTHETICS = {
    "x",
    "y",
    "alpha",
    "color",
    "colour",
    "fill",
    "linetype",
    "shape",
    "size",
    "stroke",
}

NO_GROUP = -1

# Aesthetics modifying searchers, DEPRECATED
STAT_RE = re.compile(r"\bstat\(")
DOTS_RE = re.compile(r"\.\.([a-zA-Z0-9_]+)\.\.")


class aes(Dict[str, Any]):
    """
    Create aesthetic mappings

    Parameters
    ----------
    x : str | array_like | scalar
        x aesthetic mapping
    y : str | array_like | scalar
        y aesthetic mapping
    **kwargs : dict
        Other aesthetic mappings

    Notes
    -----
    Only the **x** and **y** aesthetic mappings can be specified as
    positional arguments. All the rest must be keyword arguments.

    The value of each mapping must be one of:

    - **str**

      ```python
       import pandas as pd
       import numpy as np

       arr = [11, 12, 13]
       df = pd.DataFrame({
           "alpha": [1, 2, 3],
           "beta": [1, 2, 3],
           "gam ma": [1, 2, 3]
       })

       # Refer to a column in a dataframe
       ggplot(df, aes(x="alpha", y="beta"))
       ```

    - **array_like**

      ```python
      # A variable
      ggplot(df, aes(x="alpha", y=arr))

      # or an inplace list
      ggplot(df, aes(x="alpha", y=[4, 5, 6]))
      ```

    - **scalar**

      ```python
      # A scalar value/variable
      ggplot(df, aes(x="alpha", y=4))

      # The above statement is equivalent to
      ggplot(df, aes(x="alpha", y=[4, 4, 4]))
      ```

    - **String expression**

      ```python
      ggplot(df, aes(x="alpha", y="2*beta"))
      ggplot(df, aes(x="alpha", y="np.sin(beta)"))
      ggplot(df, aes(x="df.index", y="beta"))

      # If `count` is an aesthetic calculated by a stat
      ggplot(df, aes(x="alpha", y=after_stat("count")))
      ggplot(df, aes(x="alpha", y=after_stat("count/np.max(count)")))
      ```

      The strings in the expression can refer to;

        1. columns in the dataframe
        2. variables in the namespace
        3. aesthetic values (columns) calculated by the `stat`

      with the column names having precedence over the variables.
      For expressions, columns in the dataframe that are mapped to
      must have names that would be valid python variable names.

      This is okay:

      ```python
      # "gam ma" is a column in the dataframe
      ggplot(df, aes(x="df.index", y="gam ma"))
      ```

      While this is not:

      ```python
      # "gam ma" is a column in the dataframe, but not
      # valid python variable name
      ggplot(df, aes(x="df.index", y="np.sin(gam ma)"))
      ```

    `aes` has 2 internal methods you can use to transform variables being
    mapped.

      1. `factor` - This function turns the variable into a factor.
         It is just an alias to `pandas.Categorical`:

         ```python
         ggplot(mtcars, aes(x="factor(cyl)")) + geom_bar()
         ```

      2. `reorder` - This function changes the order of first variable
         based on values of the second variable:

         ```python
         df = pd.DataFrame({
             "x": ["b", "d", "c", "a"],
             "y": [1, 2, 3, 4]
         })

         ggplot(df, aes("reorder(x, y)", "y")) + geom_col()
         ```

    **The group aesthetic**

    `group` is a special aesthetic that the user can *map* to.
    It is used to group the plotted items. If not specified, it
    is automatically computed and in most cases the computed
    groups are sufficient. However, there may be cases were it is
    handy to map to it.

    See Also
    --------
    plotnine.after_stat : For how to map aesthetics to variable
        calculated by the stat
    plotnine.after_scale : For how to alter aesthetics after the
        data has been mapped by the scale.
    plotnine.stage : For how to map to evaluate the mapping to
        aesthetics at more than one stage of the plot building pipeline.
    """

    def __init__(self, *args, **kwargs):
        kwargs = rename_aesthetics(kwargs)
        kwargs.update(zip(("x", "y"), args))
        kwargs = self._convert_deprecated_expr(kwargs)
        self.update(kwargs)

    def __iter__(self):
        return iter(self.keys())

    def _convert_deprecated_expr(self, kwargs):
        """
        Handle old-style calculated aesthetic expression mappings

        Just converts them to use `stage` e.g.
        "stat(count)" to after_stat(count)
        "..count.." to after_stat(count)
        """
        for name, value in kwargs.items():
            if not isinstance(value, stage) and is_calculated_aes(value):
                _after_stat = strip_calculated_markers(value)
                kwargs[name] = after_stat(_after_stat)
        return kwargs

    @property
    def _starting(self) -> dict[str, Any]:
        """
        Return the subset of aesthetics mapped from the layer data

        The mapping is a dict of the form ``{name: expr}``, i.e the
        stage class has been peeled off.
        """
        d = {}
        for name, value in self.items():
            if not isinstance(value, stage):
                d[name] = value
            elif isinstance(value, stage) and value.start is not None:
                d[name] = value.start

        return d

    @property
    def _calculated(self) -> dict[str, Any]:
        """
        Return only the aesthetics mapped to calculated statistics

        The mapping is a dict of the form ``{name: expr}``, i.e the
        stage class has been peeled off.
        """
        d = {}
        for name, value in self.items():
            if isinstance(value, stage) and value.after_stat is not None:
                d[name] = value.after_stat

        return d

    @property
    def _scaled(self) -> dict[str, Any]:
        """
        Return only the aesthetics mapped to after scaling

        The mapping is a dict of the form ``{name: expr}``, i.e the
        stage class has been peeled off.
        """
        d = {}
        for name, value in self.items():
            if isinstance(value, stage) and value.after_scale is not None:
                d[name] = value.after_scale

        return d

    def __deepcopy__(self, memo):
        """
        Deep copy without copying the environment
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result

        # Just copy the keys and point to the env
        for key, item in self.items():
            result[key] = deepcopy(item, memo)

        return result

    def __radd__(self, plot):
        """
        Add aesthetic mappings to ggplot
        """
        self = deepcopy(self)
        plot.mapping.update(self)
        plot.labels.update(make_labels(self))
        return plot

    def copy(self):
        return aes(**self)

    def inherit(self, other: dict[str, Any] | aes) -> aes:
        """
        Create a  mapping that inherits aesthetics in other

        Parameters
        ----------
        other: aes | dict[str, Any]
            Default aesthetics

        Returns
        -------
        new : aes
            Aesthetic mapping
        """
        new = self.copy()
        for k in other:
            if k not in self:
                new[k] = other[k]
        return new


def rename_aesthetics(obj: THasAesNames) -> THasAesNames:
    """
    Rename aesthetics in obj

    Parameters
    ----------
    obj :
        Object that contains aesthetics names

    Returns
    -------
    :
        Object that contains aesthetics names
    """
    if isinstance(obj, dict):
        for name in tuple(obj.keys()):
            new_name = name.replace("colour", "color")
            if name != new_name:
                obj[new_name] = obj.pop(name)
    elif isinstance(obj, Sequence):
        T = type(obj)
        return T(s.replace("colour", "color") for s in obj)  # pyright: ignore
    elif obj.color is None and obj.colour is not None:
        obj.color, obj.colour = obj.colour, None

    return obj


def is_calculated_aes(ae: Any) -> bool:
    """
    Return True if Aesthetic expression maps to calculated statistic

    This function is now only used to identify the deprecated versions
    e.g. "..var.." or "stat(var)".

    Parameters
    ----------
    ae : object
        Single aesthetic mapping

    >>> is_calculated_aes("density")
    False

    >>> is_calculated_aes(4)
    False

    >>> is_calculated_aes("..density..")
    True

    >>> is_calculated_aes("stat(density)")
    True

    >>> is_calculated_aes("stat(100*density)")
    True

    >>> is_calculated_aes("100*stat(density)")
    True
    """
    if not isinstance(ae, str):
        return False
    return any(pattern.search(ae) for pattern in (STAT_RE, DOTS_RE))


def strip_stat(value):
    """
    Remove stat function that mark calculated aesthetics

    Parameters
    ----------
    value : object
        Aesthetic value. In most cases this will be a string
        but other types will pass through unmodified.

    Return
    ------
    out : object
        Aesthetic value with the dots removed.

    >>> strip_stat("stat(density + stat(count))")
    density + count

    >>> strip_stat("stat(density) + 5")
    density + 5

    >>> strip_stat("5 + stat(func(density))")
    5 + func(density)

    >>> strip_stat("stat(func(density) + var1)")
    func(density) + var1

    >>> strip_stat("stat + var1")
    stat + var1

    >>> strip_stat(4)
    4
    """

    def strip_hanging_closing_parens(s):
        """
        Remove leftover  parens
        """
        # Use and integer stack to track parens
        # and ignore leftover closing parens
        stack = 0
        idx = []
        for i, c in enumerate(s):
            if c == "(":
                stack += 1
            elif c == ")":
                stack -= 1
                if stack < 0:
                    idx.append(i)
                    stack = 0
                    continue
            yield c

    with suppress(TypeError):
        if STAT_RE.search(value):
            value = re.sub(r"\bstat\(", "", value)
            value = "".join(strip_hanging_closing_parens(value))

    return value


def strip_dots(value):
    """
    Remove dots(if any) that mark calculated aesthetics

    Parameters
    ----------
    value : object
        Aesthetic value. In most cases this will be a string
        but other types will pass through unmodified.

    Return
    ------
    out : object
        Aesthetic value with the dots removed.
    """
    with suppress(TypeError):
        value = DOTS_RE.sub(r"\1", value)
    return value


def strip_calculated_markers(value):
    """
    Remove markers for calculated aesthetics

    Parameters
    ----------
    value : object
        Aesthetic value. In most cases this will be a string
        but other types will pass through unmodified.

    Return
    ------
    out : object
        Aesthetic value with the dots removed.
    """
    return strip_stat(strip_dots(value))


def aes_to_scale(var: str):
    """
    Look up the scale that should be used for a given aesthetic
    """
    if var in {"x", "xmin", "xmax", "xend", "xintercept"}:
        var = "x"
    elif var in {"y", "ymin", "ymax", "yend", "yintercept"}:
        var = "y"
    return var


def is_position_aes(vars_: Sequence[str]):
    """
    Figure out if an aesthetic is a position aesthetic or not
    """
    return all(aes_to_scale(v) in {"x", "y"} for v in vars_)


def make_labels(mapping: dict[str, Any] | aes) -> labels_view:
    """
    Convert aesthetic mapping into text labels
    """

    def _nice_label(value: Any) -> str | None:
        if isinstance(value, str):
            return value
        elif isinstance(value, pd.Series):
            return value.name  # pyright: ignore
        elif not isinstance(value, Iterable):
            return str(value)
        elif isinstance(value, Sequence) and len(value) == 1:
            return str(value[0])
        else:
            return None

    def _make_label(ae: str, value: Any) -> str | None:
        if not isinstance(value, stage):
            return _nice_label(value)
        elif value.start is None:
            if value.after_stat is not None:
                return value.after_stat
            elif value.after_scale is not None:
                return value.after_scale
            else:
                raise ValueError("Unknown mapping")
        else:
            if value.after_stat is not None:
                return value.after_stat
            else:
                return _nice_label(value)

    valid_names = {f.name for f in fields(labels_view)}
    return labels_view(
        **{
            str(ae): _make_label(ae, label)
            for ae, label in mapping.items()
            if ae in valid_names
        }
    )


def is_valid_aesthetic(value: Any, ae: str) -> bool:
    """
    Return True if `value` looks valid.

    Parameters
    ----------
    value :
        Value to check
    ae :
        Aesthetic name

    Notes
    -----
    There are no guarantees that he value is spot on
    valid.
    """
    if ae == "linetype":
        named = {
            "solid",
            "dashed",
            "dashdot",
            "dotted",
            "_",
            "--",
            "-.",
            ":",
            "none",
            " ",
            "",
        }
        if value in named:
            return True

        # tuple of the form (offset, (on, off, on, off, ...))
        # e.g (0, (1, 2))
        conditions = [
            isinstance(value, tuple),
            isinstance(value[0], int),
            isinstance(value[1], tuple),
            len(value[1]) % 2 == 0,
            all(isinstance(x, int) for x in value[1]),
        ]
        if all(conditions):
            return True
        return False

    elif ae == "shape":
        if isinstance(value, str):
            return True

        # tuple of the form (numsides, style, angle)
        # where style is in the range [0, 3]
        # e.g (4, 1, 45)
        conditions = [
            isinstance(value, tuple),
            all(isinstance(x, int) for x in value),
            0 <= value[1] < 3,
        ]
        if all(conditions):
            return True
        return False

    elif ae in {"color", "fill"}:
        if isinstance(value, str):
            return True
        with suppress(TypeError):
            if isinstance(value, (tuple, list)) and all(
                0 <= x <= 1 for x in value
            ):
                return True
        return False

    # For any other aesthetics we return False to allow
    # for special cases to be discovered and then coded
    # for appropriately.
    return False


def has_groups(data: pd.DataFrame) -> bool:
    """
    Check if data is grouped

    Parameters
    ----------
    data :
        Data

    Returns
    -------
    out : bool
        If True, the data has groups.
    """
    # If any row in the group column is equal to NO_GROUP, then
    # the data all of them are and the data has no groups
    return data.loc[0, "group"] != NO_GROUP



import os

import numpy as np
import pandas as pd
import pytest

from plotnine import aes, annotate, geom_point, ggplot, scale_y_symlog

is_CI = os.environ.get("CI") is not None


@pytest.mark.skipif(is_CI, reason="mizani not yet shipped")
def test_scale_y_symlog():
    n = 3
    x = np.arange(-n, n + 1)
    data = pd.DataFrame({"x": x, "y": np.sign(x) * (10 ** np.abs(x))})

    p = (
        ggplot(data, aes("x", "y"))
        + geom_point()
        # Minor grid lines
        + annotate(
            "point",
            x=0,
            y=[-550, -55, -5, 5, 55, 550],
            color="red",
            shape="_",
            size=12,
        )
        + scale_y_symlog()
    )
    assert p == "scale_y_symlog"



from __future__ import annotations

import typing
from warnings import warn

import numpy as np
import pandas as pd

from .._utils import SIZE_FACTOR, order_as_data_mapping, to_rgba
from ..doctools import document
from ..exceptions import PlotnineWarning
from ..mapping import aes
from .geom import geom
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    from typing import Any

    from matplotlib.axes import Axes
    from matplotlib.offsetbox import DrawingArea

    from plotnine.coords.coord import coord
    from plotnine.iapi import panel_view
    from plotnine.layer import layer
    from plotnine.typing import DataLike


@document
class geom_vline(geom):
    """
    Vertical line

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_AES = {
        "color": "black",
        "linetype": "solid",
        "size": 0.5,
        "alpha": 1,
    }
    REQUIRED_AES = {"xintercept"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "inherit_aes": False,
    }

    legend_key_size = staticmethod(geom_segment.legend_key_size)

    def __init__(
        self,
        mapping: aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        data, mapping = order_as_data_mapping(data, mapping)
        xintercept = kwargs.pop("xintercept", None)
        if xintercept is not None:
            if mapping:
                warn(
                    "The 'xintercept' parameter has overridden "
                    "the aes() mapping.",
                    PlotnineWarning,
                )
            data = pd.DataFrame({"xintercept": np.repeat(xintercept, 1)})
            mapping = aes(xintercept="xintercept")
            kwargs["show_legend"] = False

        geom.__init__(self, mapping, data, **kwargs)

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups
        """
        ranges = coord.backtransform_range(panel_params)
        data["x"] = data["xintercept"]
        data["xend"] = data["xintercept"]
        data["y"] = ranges.y[0]
        data["yend"] = ranges.y[1]
        data = data.drop_duplicates()

        for _, gdata in data.groupby("group"):
            gdata.reset_index(inplace=True)
            geom_segment.draw_group(gdata, panel_params, coord, ax, **params)

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: layer
    ) -> DrawingArea:
        """
        Draw a vertical line in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.lines import Line2D

        x = [0.5 * da.width] * 2
        y = [0, da.height]
        data["size"] *= SIZE_FACTOR
        color = to_rgba(data["color"], data["alpha"])
        key = Line2D(
            x,
            y,
            linestyle=data["linetype"],
            linewidth=data["size"],
            color=color,
            solid_capstyle="butt",
            antialiased=False,
        )
        da.add_artist(key)
        return da



from __future__ import annotations

import typing
from copy import copy, deepcopy
from typing import Iterable, List, cast, overload

import pandas as pd

from ._utils import array_kind, check_required_aesthetics, ninteraction
from .exceptions import PlotnineError
from .mapping.aes import NO_GROUP, SCALED_AESTHETICS, aes
from .mapping.evaluation import evaluate, stage

if typing.TYPE_CHECKING:
    from typing import Any, Optional, Sequence, SupportsIndex

    from plotnine import ggplot
    from plotnine.coords.coord import coord
    from plotnine.facets.layout import Layout
    from plotnine.geoms.geom import geom
    from plotnine.layer import layer
    from plotnine.mapping import Environment
    from plotnine.positions.position import position
    from plotnine.scales.scales import Scales
    from plotnine.stats.stat import stat
    from plotnine.typing import (
        DataFrameConvertible,
        DataLike,
        LayerDataLike,
    )


class layer:
    """
    Layer

    When a `geom` or `stat` is added to a [](`~plotnine.ggplot`) object,
    it creates a single layer. This class is a representation of that layer.

    Parameters
    ----------
    geom :
        geom to used to draw this layer.
    stat :
        stat used for the statistical transformation of
        data in this layer
    mapping :
        Aesthetic mappings.
    data :
        Data plotted in this layer. If `None`, the data from
        the [](`~plotnine.ggplot`) object will be used.
    position :
        Position object to adjust the geometries in this layer.
    inherit_aes :
        If `True` inherit from the aesthetic mappings of
        the [](`~plotnine.ggplot`) object.
    show_legend :
        Whether to make up and show a legend for the mappings
        of this layer. If `None` then an automatic/good choice
        is made
    raster :
        If `True`, draw onto this layer a raster (bitmap) object
        even if the final image format is vector.

    Notes
    -----
    There is no benefit to manually creating a layer. You should
    always use a `geom` or `stat`.
    """

    # Data for this layer
    data: pd.DataFrame

    def __init__(
        self,
        geom: geom,
        stat: stat,
        *,
        mapping: aes,
        data: Optional[LayerDataLike],
        position: position,
        inherit_aes: bool = True,
        show_legend: bool | dict[str, bool] | None = None,
        raster: bool = False,
    ):
        self.geom = geom
        self.stat = stat
        self._data = data
        self.mapping = mapping
        self.position = position
        self.inherit_aes = inherit_aes
        self.show_legend = show_legend
        self.raster = raster
        self.zorder = 0

    @staticmethod
    def from_geom(geom: geom) -> layer:
        """
        Create a layer given a [](`~plotnine.geoms.geom`)

        Parameters
        ----------
        geom :
            `geom` from which a layer will be created

        Returns
        -------
        out : layer
            Layer that represents the specific `geom`.
        """
        kwargs = geom._kwargs
        lkwargs = {
            "geom": geom,
            "mapping": geom.mapping,
            "data": geom.data,
            "stat": geom._stat,
            "position": geom._position,
        }

        layer_params = ("inherit_aes", "show_legend", "raster")
        for param in layer_params:
            if param in kwargs:
                lkwargs[param] = kwargs[param]
            elif param in geom.DEFAULT_PARAMS:
                lkwargs[param] = geom.DEFAULT_PARAMS[param]
        return layer(**lkwargs)

    def __radd__(self, plot: ggplot) -> ggplot:
        """
        Add layer to ggplot object
        """
        try:
            plot.layers.append(self)
        except AttributeError as e:
            msg = f"Cannot add layer to object of type {type(plot)!r}"
            raise PlotnineError(msg) from e
        return plot

    def __deepcopy__(self, memo: dict[Any, Any]) -> layer:
        """
        Deep copy without copying the self.data dataframe
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        for key, item in old.items():
            if key == "data":
                new[key] = item
            else:
                new[key] = deepcopy(item, memo)

        return result

    def setup(self, plot: ggplot):
        """
        Prepare layer for the plot building

        Give the layer access to the data, mapping and environment
        """
        self._make_layer_data(plot.data)
        self._make_layer_mapping(plot.mapping)
        self._make_layer_environments(plot.environment)

    def _make_layer_data(self, plot_data: DataLike | None):
        """
        Generate data to be used by this layer

        Parameters
        ----------
        plot_data :
            ggplot object data
        """
        if plot_data is None:
            data = pd.DataFrame()
        elif hasattr(plot_data, "to_pandas"):
            data = cast("DataFrameConvertible", plot_data).to_pandas()
        else:
            data = cast("pd.DataFrame", plot_data)

        # Each layer that does not have data gets a copy of
        # of the ggplot.data. If it has data it is replaced
        # by copy so that we do not alter the users data
        if self._data is None:
            try:
                self.data = copy(data)
            except AttributeError as e:
                _geom_name = self.geom.__class__.__name__
                _data_name = data.__class__.__name__
                msg = (
                    f"{_geom_name} layer expects a dataframe, "
                    f"but it got {_data_name} instead."
                )
                raise PlotnineError(msg) from e
        elif callable(self._data):
            self.data = self._data(data)
            if not isinstance(self.data, pd.DataFrame):
                raise PlotnineError(
                    "Data function must return a Pandas dataframe"
                )
        else:
            # Recognise polars dataframes
            if hasattr(self._data, "to_pandas"):
                self.data = cast(
                    "DataFrameConvertible", self._data
                ).to_pandas()
            elif isinstance(self._data, pd.DataFrame):
                self.data = self._data.copy()
            else:
                raise TypeError(f"Data has a bad type: {type(self.data)}")

    def _make_layer_mapping(self, plot_mapping: aes):
        """
        Create the aesthetic mappings to be used by this layer

        Parameters
        ----------
        plot_mapping :
            ggplot object mapping
        """
        if self.inherit_aes:
            self.mapping = self.mapping.inherit(plot_mapping)

        # aesthetics set as parameters override the same
        # aesthetics set as mappings, so we can ignore
        # those in the mapping
        for ae in self.geom.aes_params:
            if ae in self.mapping:
                del self.mapping[ae]

        # Set group as a mapping if set as a parameter
        if "group" in self.geom.aes_params:
            group = self.geom.aes_params["group"]
            # Double quote str so that it evaluates to itself
            if isinstance(group, str):
                group = f'"{group}"'
            self.mapping["group"] = stage(start=group)

    def _make_layer_environments(self, plot_environment: Environment):
        """
        Create the aesthetic mappings to be used by this layer

        Parameters
        ----------
        plot_environment :
            Namespace in which to execute aesthetic expressions.
        """
        self.geom.environment = plot_environment
        self.stat.environment = plot_environment

    def compute_aesthetics(self, plot: ggplot):
        """
        Return a dataframe where the columns match the aesthetic mappings

        Transformations like 'factor(cyl)' and other
        expression evaluation are  made in here
        """
        evaled = evaluate(self.mapping._starting, self.data, plot.environment)
        evaled_aes = aes(**{str(col): col for col in evaled})
        plot.scales.add_defaults(evaled, evaled_aes)

        if len(self.data) == 0 and len(evaled) > 0:
            # No data, and vectors suppled to aesthetics
            evaled["PANEL"] = 1
        else:
            evaled["PANEL"] = self.data["PANEL"]

        data = add_group(evaled)
        self.data = data.sort_values("PANEL", kind="mergesort")

    def compute_statistic(self, layout: Layout):
        """
        Compute & return statistics for this layer
        """
        data = self.data
        if not len(data):
            return

        params = self.stat.setup_params(data)
        data = self.stat.use_defaults(data)
        data = self.stat.setup_data(data)
        data = self.stat.compute_layer(data, params, layout)
        self.data = data

    def map_statistic(self, plot: ggplot):
        """
        Mapping aesthetics to computed statistics
        """
        # Mixin default stat aesthetic mappings
        calculated = self.mapping.inherit(self.stat.DEFAULT_AES)._calculated

        if not len(self.data) or not calculated:
            return

        # The statistics are calculated in transformed space, but
        # we evaluate the mapping to them in data space.
        # NOTE: If the inverse-retransform turn out to be slow
        # we can try applying them to only the required columns.
        data = plot.scales.inverse_df(self.data)
        stat_data = evaluate(calculated, data, plot.environment)

        # If there are duplicate columns, we use the computed
        # ones in stat_data
        columns = data.columns.difference(stat_data.columns)
        data = pd.concat([data[columns], stat_data], axis=1)

        self.data = plot.scales.transform_df(data)

        # Add any new scales, if needed
        new = {ae: ae for ae in stat_data.columns}
        plot.scales.add_defaults(self.data, new)

    def setup_data(self):
        """
        Prepare/modify data for plotting
        """
        data = self.data
        if len(data) == 0:
            return

        data = self.geom.setup_data(data)

        check_required_aesthetics(
            self.geom.REQUIRED_AES,
            set(data.columns) | set(self.geom.aes_params),
            self.geom.__class__.__name__,
        )

        self.data = data

    def compute_position(self, layout: Layout):
        """
        Compute the position of each geometric object

        This is in concert with the other objects in the panel depending
        on the position class of the geom
        """
        if len(self.data) == 0:
            return

        params = self.position.setup_params(self.data)
        data = self.position.setup_data(self.data, params)
        data = self.position.compute_layer(data, params, layout)
        self.data = data

    def draw(self, layout: Layout, coord: coord):
        """
        Draw geom

        Parameters
        ----------
        layout : Layout
            Layout object created when the plot is getting
            built
        coord : coord
            Type of coordinate axes
        """
        params = copy(self.geom.params)
        params.update(self.stat.params)
        params["zorder"] = self.zorder
        params["raster"] = self.raster
        self.data = self.geom.handle_na(self.data)
        # At this point each layer must have the data
        # that is created by the plot build process
        self.geom.draw_layer(self.data, layout, coord, **params)

    def use_defaults(
        self,
        data: pd.DataFrame,
        aes_modifiers: dict[str, Any],
    ) -> pd.DataFrame:
        """
        Prepare/modify data for plotting

        Parameters
        ----------
        data :
            Data
        aes_modifiers :
            Expression to evaluate and replace aesthetics in
            the data.
        """
        return self.geom.use_defaults(data, aes_modifiers)

    def finish_statistics(self):
        """
        Prepare/modify data for plotting
        """
        self.stat.finish_layer(self.data, self.stat.params)


class Layers(List[layer]):
    """
    List of layers

    During the plot building pipeline, many operations are
    applied at all layers in the plot. This class makes those
    tasks easier.
    """

    @overload
    def __radd__(self, other: Iterable[layer]) -> Layers: ...

    @overload
    def __radd__(self, other: ggplot) -> ggplot: ...

    def __radd__(self, other: Iterable[layer] | ggplot) -> Layers | ggplot:
        """
        Add layers to ggplot object
        """
        # Add layers to ggplot object
        from .ggplot import ggplot

        if isinstance(other, ggplot):
            for obj in self:
                other += obj
        else:
            raise PlotnineError(
                f"Cannot add Layers to object of type {type(other)}"
            )
        return other

    @overload
    def __getitem__(self, key: SupportsIndex) -> layer: ...

    @overload
    def __getitem__(self, key: slice) -> Layers: ...

    def __getitem__(self, key: SupportsIndex | slice) -> layer | Layers:
        result = super().__getitem__(key)
        if isinstance(result, Iterable):
            result = Layers(result)
        return result

    @property
    def data(self) -> list[pd.DataFrame]:
        return [l.data for l in self]

    def setup(self, plot: ggplot):
        for l in self:
            l.setup(plot)

    def setup_data(self):
        for l in self:
            l.setup_data()

    def draw(self, layout: Layout, coord: coord):
        # If zorder is 0, it is left to MPL
        for i, l in enumerate(self, start=1):
            l.zorder = i
            l.draw(layout, coord)

    def compute_aesthetics(self, plot: ggplot):
        for l in self:
            l.compute_aesthetics(plot)

    def compute_statistic(self, layout: Layout):
        for l in self:
            l.compute_statistic(layout)

    def map_statistic(self, plot: ggplot):
        for l in self:
            l.map_statistic(plot)

    def compute_position(self, layout: Layout):
        for l in self:
            l.compute_position(layout)

    def use_defaults_after_scale(self):
        for l in self:
            l.use_defaults(l.data, l.mapping._scaled)

    def transform(self, scales: Scales):
        for l in self:
            l.data = scales.transform_df(l.data)

    def train(self, scales: Scales):
        for l in self:
            scales.train_df(l.data)

    def map(self, scales: Scales):
        for l in self:
            l.data = scales.map_df(l.data)

    def finish_statistics(self):
        for l in self:
            l.finish_statistics()

    def update_labels(self, plot: ggplot):
        for l in self:
            plot._update_labels(l)


def add_group(data: pd.DataFrame) -> pd.DataFrame:
    """
    Add group to the dataframe

    The group depends on the interaction of the discrete
    aesthetic columns in the dataframe.
    """
    if len(data) == 0:
        return data

    if "group" not in data:
        ignore = data.columns.difference(list(SCALED_AESTHETICS))
        disc = discrete_columns(data, ignore=ignore)
        if disc:
            data["group"] = ninteraction(data[disc], drop=True)
        else:
            data["group"] = NO_GROUP
    else:
        data["group"] = ninteraction(data[["group"]], drop=True)

    return data


def discrete_columns(
    df: pd.DataFrame, ignore: Sequence[str] | pd.Index
) -> list[str]:
    """
    Return a list of the discrete columns in the dataframe

    Parameters
    ----------
    df : dataframe
        Data
    ignore : list[str]
        A list|set|tuple with the names of the columns to skip.
    """
    lst = []
    for col in df:
        if array_kind.discrete(df[col]) and (col not in ignore):
            # Some columns are represented as object dtype
            # but may have compound structures as values.
            try:
                hash(df[col].iloc[0])
            except TypeError:
                continue
            lst.append(str(col))
    return lst



from __future__ import annotations

from abc import ABCMeta
from collections import defaultdict
from typing import TYPE_CHECKING
from weakref import WeakValueDictionary

from ..exceptions import PlotnineError

if TYPE_CHECKING:
    from typing import TypeVar

    T = TypeVar("T")


def alias(cls: type[T]) -> type[T]:
    """
    Add docstring that class is an alias of its base class
    """
    if cls.__doc__ is not None:
        return cls

    base = cls.__bases__[0]
    name = base.__name__
    qualname = f"{base.__module__}.{name}"
    cls.__doc__ = f"alias of [{name}](`{qualname}`)"
    return cls


class _Registry(WeakValueDictionary):
    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError as e:
            msg = (
                f"'{key}' Not in Registry. Make sure the module in "
                "which it is defined has been imported."
            )
            raise PlotnineError(msg) from e


Registry = _Registry()


class Register(ABCMeta):
    """
    Creates class that automatically registers all subclasses

    To prevent the base class from showing up in the registry,
    it should inherit from ABC. This metaclass uses a single
    dictionary to register all types of subclasses.

    To access the registered objects, use:

        obj = Registry['name']

    To explicitly register objects

        Registry['name'] = obj

    Notes
    -----
    When objects are deleted, they are automatically removed
    from the Registry.
    """

    # namespace is of the class that subclasses Registry (or a
    # subclasses a subclass of Registry, ...) being created
    # e.g. geom, geom_point, ...
    def __new__(cls, name, bases, namespace):
        sub_cls = super().__new__(cls, name, bases, namespace)
        is_base_class = len(bases) and bases[0].__name__ == "ABC"
        if not is_base_class:
            Registry[name] = sub_cls
        return sub_cls


class RegistryHierarchyMeta(type):
    """
    Create a class that registers subclasses and the Hierarchy

    The class has gets two properties:

    1. `_registry` a dictionary of all the subclasses of the
       base class. The keys are the names of the classes and
       the values are the class objects.
    2. `_hierarchy` a dictionary (default) that holds the
       inheritance hierarchy of each class. Each key is a class
       and the value is a list of classes. The first name in the
       list is that of the key class.

    The goal of the `_hierarchy` object to facilitate the
    lookup of themeable properties taking into consideration the
    inheritance hierarchy. For example if `strip_text_x` inherits
    from `strip_text` which inherits from `text`, then if a property
    of `strip_text_x` is requested, the lookup should fallback to
    the other two if `strip_text_x` is not present or is missing
    the requested property.
    """

    def __init__(cls, name, bases, namespace):
        if not hasattr(cls, "_registry"):
            cls._registry = {}
            cls._hierarchy = defaultdict(list)
        else:
            cls._registry[name] = cls
            cls._hierarchy[name].append(name)
            for base in bases:
                for base2 in base.mro()[:-2]:
                    cls._hierarchy[base2.__name__].append(name)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, fields
from functools import cached_property
from typing import TYPE_CHECKING, Literal, cast
from warnings import warn

import numpy as np
import pandas as pd

from .._utils import ensure_xy_location
from .._utils.registry import Registry
from ..exceptions import PlotnineError, PlotnineWarning
from ..iapi import (
    inside_legend,
    legend_artists,
    legend_justifications_view,
    outside_legend,
)
from ..mapping.aes import rename_aesthetics
from .guide import guide

if TYPE_CHECKING:
    from typing import Literal, Optional, Sequence

    from matplotlib.offsetbox import OffsetBox, PackerBase

    from plotnine import ggplot, theme
    from plotnine.iapi import labels_view
    from plotnine.scales.scale import scale
    from plotnine.scales.scales import Scales
    from plotnine.typing import (
        LegendOnly,
        LegendOrColorbar,
        LegendPosition,
        NoGuide,
        Orientation,
        ScaledAestheticsName,
        SidePosition,
        TextJustification,
        TupleFloat2,
    )


# Terminology
# -----------
# - A guide is either a legend or colorbar.
#
# - A guide definition (gdef) is an instantiated guide as it
#   is used in the process of creating the legend
#
# - The guides class holds all guides that will appear in the
#   plot
#
# - A guide box is a fully drawn out guide.
#   It is of subclass matplotlib.offsetbox.Offsetbox


@dataclass
class guides:
    """
    Guides for each scale

    Used to assign or remove a particular guide to the scale
    of an aesthetic.
    """

    alpha: Optional[LegendOrColorbar | NoGuide] = None
    """Guide for alpha scale."""

    color: Optional[LegendOrColorbar | NoGuide] = None
    """Guide for color scale."""

    fill: Optional[LegendOrColorbar | NoGuide] = None
    """Guide for fill scale."""

    linetype: Optional[LegendOnly | NoGuide] = None
    """Guide for linetype scale."""

    shape: Optional[LegendOnly | NoGuide] = None
    """Guide for shape scale."""

    size: Optional[LegendOnly | NoGuide] = None
    """Guide for size scale."""

    stroke: Optional[LegendOnly | NoGuide] = None
    """Guide for stroke scale."""

    colour: Optional[LegendOnly | NoGuide] = None
    """Guide for colour scale."""

    def __post_init__(self):
        self.plot: ggplot
        self.plot_scales: Scales
        self.plot_labels: labels_view
        self.elements: GuidesElements
        self._lookup: dict[tuple[scale, ScaledAestheticsName], guide] = {}
        if self.colour is not None and self.color is not None:
            raise ValueError("Got a guide for color and colour, choose one.")
        rename_aesthetics(self)

    def __radd__(self, plot: ggplot):
        """
        Add guides to the plot

        Parameters
        ----------
        plot :
            ggplot object being created

        Returns
        -------
        :
            ggplot object with guides.
        """
        for f in fields(self):
            if (g := getattr(self, f.name)) is not None:
                setattr(plot.guides, f.name, g)

        return plot

    def _build(self) -> Sequence[guide]:
        """
        Build the guides

        Returns
        -------
        :
            The individual guides for which the geoms that draw them have
            have been created.
        """
        return self._create_geoms(self._merge(self._train()))

    def _setup(self, plot: ggplot):
        """
        Setup all guides that will be active
        """
        self.plot = plot
        self.elements = GuidesElements(self.plot.theme)

        guide_lookup = {
            f.name: g
            for f in fields(self)
            if (g := getattr(self, f.name)) is not None
        }

        for scale in self.plot.scales:
            for ae in scale.aesthetics:
                # The guide for aesthetic 'xxx' is stored
                # in plot.guides['xxx']. The priority for
                # the guides depends on how they are created
                # 1. ... + guides(xxx=guide_blah())
                # 2. ... + scale_xxx(guide=guide_blah())
                # 3. default(either guide_legend or guide_colorbar
                #            depending on the scale type)
                # ae = scale.aesthetics[0]

                # No guide
                if (g := guide_lookup.get(ae)) in ("none", False) or (
                    g is None and (g := scale.guide) is None
                ):
                    continue

                # check the validity of guide.
                # if guide is str, then find the guide object
                if isinstance(g, str):
                    g = Registry[f"guide_{g}"]()
                elif not isinstance(g, guide):
                    raise PlotnineError(f"Unknown guide: {g}")

                g.setup(self)
                self._lookup[(scale, ae)] = g

    def _train(self) -> Sequence[guide]:
        """
        Compute all the required guides

        Returns
        -------
        gdefs : list
            Guides for the plots
        """
        gdefs: list[guide] = []
        for (scale, ae), g in self._lookup.items():
            # Guide turned off
            if not g.elements.position:
                continue

            # check the consistency of the guide and scale.
            if (
                "any" not in g.available_aes
                and scale.aesthetics[0] not in g.available_aes
            ):
                raise PlotnineError(
                    f"{g.__class__.__name__} cannot be used for "
                    f"{scale.aesthetics}"
                )

            # title
            if g.title is None:
                if scale.name:
                    g.title = scale.name
                else:
                    g.title = getattr(self.plot.labels, ae)
                    if g.title is None:
                        warn(
                            f"Cannot generate legend for the {ae!r} "
                            "aesthetic. Make sure you have mapped a "
                            "variable to it",
                            PlotnineWarning,
                        )

            # each guide object trains scale within the object,
            # so Guides (i.e., the container of guides)
            # need not to know about them
            g = g.train(scale, ae)

            if g is not None:
                gdefs.append(g)

        return gdefs

    def _merge(self, gdefs: Sequence[guide]) -> Sequence[guide]:
        """
        Merge overlapped guides

        For example:

        ```python
         from plotnine import *
         p = (
            ggplot(mtcars, aes(y="wt", x="mpg", colour="factor(cyl)"))
            + stat_smooth(aes(fill="factor(cyl)"), method="lm")
            + geom_point()
         )
        ```

        would create two guides with the same hash
        """
        if not gdefs:
            return []

        # group guide definitions by hash, and
        # reduce each group to a single guide
        # using the guide.merge method
        definitions = pd.DataFrame(
            {"gdef": gdefs, "hash": [g.hash for g in gdefs]}
        )
        grouped = definitions.groupby("hash", sort=False)
        gdefs = []
        for name, group in grouped:
            # merge
            gdef = group["gdef"].iloc[0]
            for g in group["gdef"].iloc[1:]:
                gdef = gdef.merge(g)
            gdefs.append(gdef)
        return gdefs

    def _create_geoms(
        self,
        gdefs: Sequence[guide],
    ) -> Sequence[guide]:
        """
        Add geoms to the guide definitions
        """
        return [_g for g in gdefs if (_g := g.create_geoms())]

    def _apply_guide_themes(self, gdefs: list[guide]):
        """
        Apply the theme for each guide
        """
        for g in gdefs:
            g.theme.apply()

    def _assemble_guides(
        self,
        gdefs: list[guide],
        boxes: list[PackerBase],
    ) -> legend_artists:
        """
        Assemble guides into Anchored Offset boxes depending on location
        """
        from matplotlib.font_manager import FontProperties
        from matplotlib.offsetbox import HPacker, VPacker

        from .._mpl.offsetbox import FlexibleAnchoredOffsetbox

        elements = self.elements

        # Combine all the guides into a single box
        # The direction matters only when there is more than legend
        lookup: dict[Orientation, type[PackerBase]] = {
            "horizontal": HPacker,
            "vertical": VPacker,
        }

        def _anchored_offset_box(boxes: list[PackerBase]):
            """
            Put a group of guides into a single box for drawing
            """
            packer = lookup[elements.box]

            box = packer(
                children=boxes,  # type: ignore
                align=elements.box_just,
                pad=elements.box_margin,
                sep=elements.spacing,
            )

            return FlexibleAnchoredOffsetbox(
                xy_loc=(0.5, 0.5),
                child=box,
                pad=1,
                frameon=False,
                prop=FontProperties(size=0, stretch=0),
                bbox_to_anchor=(0, 0),
                bbox_transform=self.plot.figure.transFigure,
                borderpad=0.0,
                zorder=99.1,
            )

        # Group together guides for each position
        groups: dict[
            tuple[SidePosition, float] | tuple[TupleFloat2, TupleFloat2],
            list[PackerBase],
        ] = defaultdict(list)

        for g, b in zip(gdefs, boxes):
            groups[g._resolved_position_justification].append(b)

        legends = legend_artists()

        # Create an anchoredoffsetbox for each group/position
        for (position, just), group in groups.items():
            aob = _anchored_offset_box(group)
            if isinstance(position, str) and isinstance(just, (float, int)):
                setattr(legends, position, outside_legend(aob, just))
            else:
                position = cast(tuple[float, float], position)
                just = cast(tuple[float, float], just)
                legends.inside.append(inside_legend(aob, just, position))

        return legends

    def draw(self) -> Optional[OffsetBox]:
        """
        Draw guides onto the figure

        Returns
        -------
        :matplotlib.offsetbox.Offsetbox | None
            A box that contains all the guides for the plot.
            If there are no guides, **None** is returned.
        """
        if self.elements.position == "none":
            return

        if not (gdefs := self._build()):
            return

        # Order of guides
        # 0 do not sort, any other sorts
        # place the guides according to the guide.order
        default = max(g.order for g in gdefs) + 1
        orders = [default if g.order == 0 else g.order for g in gdefs]
        idx: list[int] = list(np.argsort(orders))
        gdefs = [gdefs[i] for i in idx]

        # Draw each guide into a box
        # Because we can have more than one guide, we keep record of
        # the drawn artists using lists
        guide_boxes = [g.draw() for g in gdefs]

        self._apply_guide_themes(gdefs)
        legends = self._assemble_guides(gdefs, guide_boxes)
        for aob in legends.boxes:
            self.plot.figure.add_artist(aob)

        self.plot.theme.targets.legends = legends


VALID_JUSTIFICATION_WORDS = {"left", "right", "top", "bottom", "center"}


@dataclass
class GuidesElements:
    """
    Theme elements used when assembling the guides object

    This class is meant to provide convenient access to all the required
    elements having worked out good defaults for the unspecified values.
    """

    theme: theme

    @cached_property
    def box(self) -> Orientation:
        """
        The direction to layout the guides
        """
        if not (box := self.theme.getp("legend_box")):
            box = (
                "vertical"
                if self.position in {"left", "right"}
                else "horizontal"
            )
        return box

    @cached_property
    def position(self) -> LegendPosition | Literal["none"]:
        if (pos := self.theme.getp("legend_position", "right")) == "inside":
            pos = self._position_inside
        return pos

    @cached_property
    def _position_inside(self) -> LegendPosition:
        # We return the position inside the panels when it is explicitly
        # set. Otherwise we return the justification inside the panels.
        # We convert the string (left, right, ...) justifications into
        # locations which justify the legend along the edges of the panel
        # area.
        # Overall when only the inside position is set, the same value is
        # applied to the justification and vice-versa. Always defaulting
        # to a center justification can have the legends close to the
        # edge go out of bounds.
        pos = self.theme.getp("legend_position_inside")
        if isinstance(pos, tuple):
            return pos

        just = self.theme.getp("legend_justification_inside", (0.5, 0.5))
        return ensure_xy_location(just)

    @cached_property
    def box_just(self) -> TextJustification:
        if not (box_just := self.theme.getp("legend_box_just")):
            box_just = (
                "left" if self.position in {"left", "right"} else "right"
            )
        return box_just

    @cached_property
    def box_margin(self) -> int:
        return self.theme.getp("legend_box_margin")

    @cached_property
    def spacing(self) -> float:
        return self.theme.getp("legend_spacing")

    @cached_property
    def justification(self) -> legend_justifications_view:
        # Don't bother, the legend has been turned off
        if self.position == "none":
            return legend_justifications_view()

        dim_lookup = {"left": 1, "right": 1, "top": 0, "bottom": 0}

        # Process justification for legend on left, right, top & bottom
        def _lrtb(pos):
            just = self.theme.getp(f"legend_justification_{pos}")
            idx = dim_lookup[pos]
            if just is None:
                just = (0.5, 0.5)
            elif just in VALID_JUSTIFICATION_WORDS:
                just = ensure_xy_location(just)
            elif isinstance(just, (float, int)):
                just = (just, just)
            return just[idx]

        # Process justification for legend inside the panels
        def _inside():
            just = self.theme.getp("legend_justification_inside")
            if just is None:
                return None
            return ensure_xy_location(just)

        # For legends outside the panels, the default justification is
        # the center (0.5). For legends inside the panels, the default
        # is the position. The final value of the position comes from
        # the guide._final_position.
        return legend_justifications_view(
            left=_lrtb("left"),
            right=_lrtb("right"),
            top=_lrtb("top"),
            bottom=_lrtb("bottom"),
            inside=_inside(),
        )
